#
# Copyright (C) 2007 Andrew Resch <andrewresch@gmail.com>
# Copyright (C) 2010 Pedro Algarvio <pedro@algarvio.me>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#
import os
import sys
from logging import DEBUG, FileHandler, getLogger

from twisted.internet.error import CannotListenError

from deluge.argparserbase import ArgParserBase
from deluge.common import run_profiled
from deluge.configmanager import get_config_dir
from deluge.i18n import setup_mock_translation


def add_daemon_options(parser):
    group = parser.add_argument_group(_('Daemon Options'))
    group.add_argument(
        '-u',
        '--ui-interface',
        metavar='<ip-addr>',
        action='store',
        help=_('IP address to listen for UI connections'),
    )
    group.add_argument(
        '-p',
        '--port',
        metavar='<port>',
        action='store',
        type=int,
        help=_('Port to listen for UI connections on'),
    )
    group.add_argument(
        '-i',
        '--interface',
        metavar='<ip-addr>',
        dest='listen_interface',
        action='store',
        help=_('IP address to listen for BitTorrent connections'),
    )
    group.add_argument(
        '-o',
        '--outgoing-interface',
        metavar='<interface>',
        dest='outgoing_interface',
        action='store',
        help=_(
            'The network interface name or IP address for outgoing BitTorrent connections.'
        ),
    )
    group.add_argument(
        '--read-only-config-keys',
        metavar='<comma-separated-keys>',
        action='store',
        help=_('Config keys to be unmodified by `set_config` RPC'),
        type=str,
        default='',
    )
    parser.add_process_arg_group()


def start_daemon(skip_start=False):
    """
    Entry point for daemon script

    Args:
        skip_start (bool): If starting daemon should be skipped.

    Returns:
        deluge.core.daemon.Daemon: A new daemon object

    """
    setup_mock_translation()

    # Setup the argument parser
    parser = ArgParserBase()
    add_daemon_options(parser)

    options = parser.parse_args()

    # Check for any daemons running with this same config
    from deluge.core.daemon import is_daemon_running

    pid_file = get_config_dir('deluged.pid')
    if is_daemon_running(pid_file):
        print(
            'Cannot run multiple daemons with same config directory.\n'
            'If you believe this is an error, force starting by deleting: %s' % pid_file
        )
        sys.exit(1)

    log = getLogger(__name__)

    # If no logfile specified add logging to default location (as well as stdout)
    if not options.logfile:
        options.logfile = get_config_dir('deluged.log')
        file_handler = FileHandler(options.logfile)
        log.addHandler(file_handler)

    def run_daemon(options):
        try:
            from deluge.core.daemon import Daemon

            daemon = Daemon(
                listen_interface=options.listen_interface,
                outgoing_interface=options.outgoing_interface,
                interface=options.ui_interface,
                port=options.port,
                read_only_config_keys=options.read_only_config_keys.split(','),
            )
            if skip_start:
                return daemon
            else:
                daemon.start()
        except CannotListenError as ex:
            log.error(
                'Cannot start deluged, listen port in use.\n'
                ' Check for other running daemons or services using this port: %s:%s',
                ex.interface,
                ex.port,
            )
            sys.exit(1)
        except Exception as ex:
            log.error('Unable to start deluged: %s', ex)
            if log.isEnabledFor(DEBUG):
                log.exception(ex)
            sys.exit(1)
        finally:
            log.info('Exiting...')
            if options.pidfile:
                os.remove(options.pidfile)

    return run_profiled(
        run_daemon, options, output_file=options.profile, do_profile=options.profile
    )



#
# Copyright (C) 2008 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging
from xml.sax.saxutils import escape as xml_escape

import deluge.component as component
from deluge.common import decode_bytes, fdate, fsize, is_url

from .tab_data_funcs import fdate_or_dash, fpieces_num_size
from .torrentdetails import Tab

log = logging.getLogger(__name__)


class DetailsTab(Tab):
    def __init__(self):
        super().__init__('Details', 'details_tab', 'details_tab_label')

        self.add_tab_widget('summary_name', None, ('name',))
        self.add_tab_widget('summary_total_size', fsize, ('total_size',))
        self.add_tab_widget('summary_num_files', str, ('num_files',))
        self.add_tab_widget('summary_completed', fdate_or_dash, ('completed_time',))
        self.add_tab_widget('summary_date_added', fdate, ('time_added',))
        self.add_tab_widget('summary_torrent_path', None, ('download_location',))
        self.add_tab_widget('summary_hash', str, ('hash',))
        self.add_tab_widget('summary_comments', str, ('comment',))
        self.add_tab_widget('summary_creator', str, ('creator',))
        self.add_tab_widget(
            'summary_pieces', fpieces_num_size, ('num_pieces', 'piece_length')
        )

    def update(self):
        # Get the first selected torrent
        selected = component.get('TorrentView').get_selected_torrents()

        # Only use the first torrent in the list or return if None selected
        if selected:
            selected = selected[0]
        else:
            # No torrent is selected in the torrentview
            self.clear()
            return

        session = component.get('SessionProxy')
        session.get_torrent_status(selected, self.status_keys).addCallback(
            self._on_get_torrent_status
        )

    def _on_get_torrent_status(self, status):
        # Check to see if we got valid data from the core
        if status is None:
            return

        # Update all the label widgets
        for widget in self.tab_widgets.values():
            txt = xml_escape(self.widget_status_as_fstr(widget, status))
            if decode_bytes(widget.obj.get_text()) != txt:
                if 'comment' in widget.status_keys and is_url(txt):
                    widget.obj.set_markup(f'<a href="{txt}">{txt}</a>')
                else:
                    widget.obj.set_markup(txt)

    def clear(self):
        for widget in self.tab_widgets.values():
            widget.obj.set_text('')



#
# Copyright (C) 2007 Marcos Mobley ('markybob') <markybob@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

from datetime import date

from gi.repository import Gtk

import deluge.component as component
from deluge.common import get_version, open_url_in_browser, windows_check
from deluge.ui.client import client

from .common import get_deluge_icon, get_pixbuf


class AboutDialog:
    def __init__(self):
        self.about = Gtk.AboutDialog()
        self.about.set_transient_for(component.get('MainWindow').window)
        self.about.set_position(Gtk.WindowPosition.CENTER)
        self.about.set_name(_('Deluge'))
        self.about.set_program_name(_('Deluge'))
        if windows_check():

            def url_hook(dialog, url):
                """Url hook for Windows OS which has no default browser."""
                open_url_in_browser(url)
                return True

            self.about.connect('activate-link', url_hook)

        version = get_version()

        self.about.set_copyright(
            _('Copyright %(year_start)s-%(year_end)s Deluge Team')
            % {'year_start': 2007, 'year_end': date.today().year}
        )
        self.about.set_comments(
            _('A peer-to-peer file sharing program\nutilizing the BitTorrent protocol.')
            + '\n\n'
            + _('Client:')
            + ' %s\n' % version
        )
        self.about.set_version(version)
        self.about.set_authors(
            [
                _('Current Developers:'),
                'Andrew Resch',
                'Damien Churchill',
                'John Garland',
                'Calum Lind',
                '',
                'libtorrent (libtorrent.org):',
                'Arvid Norberg',
                '',
                _('Past Developers or Contributors:'),
                'Zach Tibbitts',
                'Alon Zakai',
                'Marcos Mobley',
                'Alex Dedul',
                'Sadrul Habib Chowdhury',
                'Ido Abramovich',
                'Martijn Voncken',
            ]
        )
        self.about.set_artists(['Andrew Wedderburn', 'Andrew Resch'])
        self.about.set_translator_credits(
            '\n'.join(
                [
                    'Aaron Wang Shi',
                    'abbigss',
                    'ABCdatos',
                    'Abcx',
                    'Actam',
                    'Adam',
                    'adaminikisi',
                    'adi_oporanu',
                    'Adrian Goll',
                    'afby',
                    'Ahmades',
                    'Ahmad Farghal',
                    'Ahmad Gharbeia أحمد غربية',
                    'akira',
                    'Aki Sivula',
                    'Alan Pepelko',
                    'Alberto',
                    'Alberto Ferrer',
                    'alcatr4z',
                    'AlckO',
                    'Aleksej Korgenkov',
                    'Alessio Treglia',
                    'Alexander Ilyashov',
                    'Alexander Matveev',
                    'Alexander Saltykov',
                    'Alexander Taubenkorb',
                    'Alexander Telenga',
                    'Alexander Yurtsev',
                    'Alexandre Martani',
                    'Alexandre Rosenfeld',
                    'Alexandre Sapata Carbonell',
                    'Alexey Osipov',
                    'Alin Claudiu Radut',
                    'allah',
                    'AlSim',
                    'Alvaro Carrillanca P.',
                    'A.Matveev',
                    'Andras Hipsag',
                    'András Kárász',
                    'Andrea Ratto',
                    'Andreas Johansson',
                    'Andreas Str',
                    'André F. Oliveira',
                    'AndreiF',
                    'andrewh',
                    'Angel Guzman Maeso',
                    'Aníbal Deboni Neto',
                    'animarval',
                    'Antonio Cono',
                    'antoniojreyes',
                    'Anton Shestakov',
                    'Anton Yakutovich',
                    'antou',
                    'Arkadiusz Kalinowski',
                    'Artin',
                    'artir',
                    'Astur',
                    'Athanasios Lefteris',
                    'Athmane MOKRAOUI (ButterflyOfFire)',
                    'Augusta Carla Klug',
                    'Avoledo Marco',
                    'axaard',
                    'AxelRafn',
                    'Axezium',
                    'Ayont',
                    'b3rx',
                    'Bae Taegil',
                    'Bajusz Tamás',
                    "Balaam's Miracle",
                    'Ballestein',
                    'Bent Ole Fosse',
                    'berto89',
                    'bigx',
                    'Bjorn Inge Berg',
                    'blackbird',
                    'Blackeyed',
                    'blackmx',
                    'BlueSky',
                    'Blutheo',
                    'bmhm',
                    'bob00work',
                    'boenki',
                    'Bogdan Bădic-Spătariu',
                    'bonpu',
                    'Boone',
                    'boss01',
                    'Branislav Jovanović',
                    'bronze',
                    'brownie',
                    'Brus46',
                    'bumper',
                    'butely',
                    'BXCracer',
                    'c0nfidencal',
                    'Can Kaya',
                    'Carlos Alexandro Becker',
                    'cassianoleal',
                    'Cédric.h',
                    'César Rubén',
                    'chaoswizard',
                    'Chen Tao',
                    'chicha',
                    'Chien Cheng Wei',
                    'Christian Kopac',
                    'Christian Widell',
                    'Christoffer Brodd-Reijer',
                    'christooss',
                    'CityAceE',
                    'Clopy',
                    'Clusty',
                    'cnu',
                    'Commandant',
                    'Constantinos Koniaris',
                    'Coolmax',
                    'cosmix',
                    'Costin Chirvasuta',
                    'CoVaLiDiTy',
                    'cow_2001',
                    'Crispin Kirchner',
                    'crom',
                    'Cruster',
                    'Cybolic',
                    'Dan Bishop',
                    'Danek',
                    'Dani',
                    'Daniel Demarco',
                    'Daniel Ferreira',
                    'Daniel Frank',
                    'Daniel Holm',
                    'Daniel Høyer Iversen',
                    'Daniel Marynicz',
                    'Daniel Nylander',
                    'Daniel Patriche',
                    'Daniel Schildt',
                    'Daniil Sorokin',
                    'Dante Díaz',
                    'Daria Michalska',
                    'DarkenCZ',
                    'Darren',
                    'Daspah',
                    'David Eurenius',
                    'davidhjelm',
                    'David Machakhelidze',
                    'Dawid Dziurdzia',
                    'Daya Adianto ',
                    'dcruz',
                    'Deady',
                    'Dereck Wonnacott',
                    'Devgru',
                    'Devid Antonio Filoni' 'DevilDogTG',
                    'di0rz`',
                    'Dialecti Valsamou',
                    'Diego Medeiros',
                    'Dkzoffy',
                    'Dmitrij D. Czarkoff',
                    'Dmitriy Geels',
                    'Dmitry Olyenyov',
                    'Dominik Kozaczko',
                    'Dominik Lübben',
                    'doomster',
                    'Dorota Król',
                    'Doyen Philippe',
                    'Dread Knight',
                    'DreamSonic',
                    'duan',
                    'Duong Thanh An',
                    'DvoglavaZver',
                    'dwori',
                    'dylansmrjones',
                    'Ebuntor',
                    'Edgar Alejandro Jarquin Flores',
                    'Eetu',
                    'ekerazha',
                    'Elias Julkunen',
                    'elparia',
                    'Emberke',
                    'Emiliano Goday Caneda',
                    'EndelWar',
                    'eng.essam',
                    'enubuntu',
                    'ercangun',
                    'Erdal Ronahi',
                    'ergin üresin',
                    'Eric',
                    'Éric Lassauge',
                    'Erlend Finvåg',
                    'Errdil',
                    'ethan shalev',
                    'Evgeni Spasov',
                    'ezekielnin',
                    'Fabian Ordelmans',
                    'Fabio Mazanatti',
                    'Fábio Nogueira',
                    'FaCuZ',
                    'Felipe Lerena',
                    'Fernando Pereira',
                    'fjetland',
                    'Florian Schäfer',
                    'FoBoS',
                    'Folke',
                    'Force',
                    'fosk',
                    'fragarray',
                    'freddeg',
                    'Frédéric Perrin',
                    'Fredrik Kilegran',
                    'FreeAtMind',
                    'Fulvio Ciucci',
                    'Gabor Kelemen',
                    'Galatsanos Panagiotis',
                    'Gaussian',
                    'gdevitis',
                    'Georg Brzyk',
                    'George Dumitrescu',
                    'Georgi Arabadjiev',
                    'Georg Sieber',
                    'Gerd Radecke',
                    'Germán Heusdens',
                    'Gianni Vialetto',
                    'Gigih Aji Ibrahim',
                    'Giorgio Wicklein',
                    'Giovanni Rapagnani',
                    'Giuseppe',
                    'gl',
                    'glen',
                    'granjerox',
                    'Green Fish',
                    'greentea',
                    'Greyhound',
                    'G. U.',
                    'Guillaume BENOIT',
                    'Guillaume Pelletier',
                    'Gustavo Henrique Klug',
                    'gutocarvalho',
                    'Guybrush88',
                    'Hans Rødtang',
                    'HardDisk',
                    'Hargas Gábor',
                    'Heitor Thury Barreiros Barbosa',
                    'helios91940',
                    'helix84',
                    'Helton Rodrigues',
                    'Hendrik Luup',
                    'Henrique Ferreiro',
                    'Henry Goury-Laffont',
                    'Hezy Amiel',
                    'hidro',
                    'hoball',
                    'hokten',
                    'Holmsss',
                    'hristo.num',
                    'Hubert Życiński',
                    'Hyo',
                    'Iarwain',
                    'ibe',
                    'ibear',
                    'Id2ndR',
                    'Igor Zubarev',
                    'IKON (Ion)',
                    'imen',
                    'Ionuț Jula',
                    'Isabelle STEVANT',
                    'István Nyitrai',
                    'Ivan Petrovic',
                    'Ivan Prignano',
                    'IvaSerge',
                    'jackmc',
                    'Jacks0nxD',
                    'Jack Shen',
                    'Jacky Yeung',
                    'Jacques Stadler',
                    'Janek Thomaschewski',
                    'Jan Kaláb',
                    'Jan Niklas Hasse',
                    'Jasper Groenewegen',
                    'Javi Rodríguez',
                    'Jayasimha (ಜಯಸಿಂಹ)',
                    'jeannich',
                    'Jeff Bailes',
                    'Jesse Zilstorff',
                    'Joan Duran',
                    'João Santos',
                    'Joar Bagge',
                    'Joe Anderson',
                    'Joel Calado',
                    'Johan Linde',
                    'John Garland',
                    'Jojan',
                    'jollyr0ger',
                    'Jonas Bo Grimsgaard',
                    'Jonas Granqvist',
                    'Jonas Slivka',
                    'Jonathan Zeppettini',
                    'Jørgen',
                    'Jørgen Tellnes',
                    'josé',
                    'José Geraldo Gouvêa',
                    'José Iván León Islas',
                    'José Lou C.',
                    'Jose Sun',
                    'Jr.',
                    'Jukka Kauppinen',
                    'Julián Alarcón',
                    'julietgolf',
                    'Jusic',
                    'Justzupi',
                    'Kaarel',
                    'Kai Thomsen',
                    'Kalman Tarnay',
                    'Kamil Páral',
                    'Kane_F',
                    'kaotiks@gmail.com',
                    'Kateikyoushii',
                    'kaxhinaz',
                    'Kazuhiro NISHIYAMA',
                    'Kerberos',
                    'Keresztes Ákos',
                    'kevintyk',
                    'kiersie',
                    'Kimbo^',
                    'Kim Lübbe',
                    'kitzOgen',
                    'Kjetil Rydland',
                    'kluon',
                    'kmikz',
                    'Knedlyk',
                    'koleoptero',
                    'Kőrösi Krisztián',
                    'Kouta',
                    'Krakatos',
                    'Krešo Kunjas',
                    'kripken',
                    'Kristaps',
                    'Kristian Øllegaard',
                    'Kristoffer Egil Bonarjee',
                    'Krzysztof Janowski',
                    'Krzysztof Zawada',
                    'Larry Wei Liu',
                    'laughterwym',
                    'Laur Mõtus',
                    'lazka',
                    'leandrud',
                    'lê bình',
                    'Le Coz Florent',
                    'Leo',
                    'liorda',
                    'LKRaider',
                    'LoLo_SaG',
                    'Long Tran',
                    'Lorenz',
                    'Low Kian Seong',
                    'Luca Andrea Rossi',
                    'Luca Ferretti',
                    'Lucky LIX',
                    'Luis Gomes',
                    'Luis Reis',
                    'Łukasz Wyszyński',
                    'luojie-dune',
                    'maaark',
                    'Maciej Chojnacki',
                    'Maciej Meller',
                    'Mads Peter Rommedahl',
                    'Major Kong',
                    'Malaki',
                    'malde',
                    'Malte Lenz',
                    'Mantas Kriaučiūnas',
                    'Mara Sorella',
                    'Marcin',
                    'Marcin Falkiewicz',
                    'marcobra',
                    'Marco da Silva',
                    'Marco de Moulin',
                    'Marco Rodrigues',
                    'Marcos',
                    'Marcos Escalier',
                    'Marcos Mobley',
                    'Marcus Ekstrom',
                    'Marek Dębowski',
                    'Mário Buči',
                    'Mario Munda',
                    'Marius Andersen',
                    'Marius Hudea',
                    'Marius Mihai',
                    'Mariusz Cielecki',
                    'Mark Krapivner',
                    'marko-markovic',
                    'Markus Brummer',
                    'Markus Sutter',
                    'Martin',
                    'Martin Dybdal',
                    'Martin Iglesias',
                    'Martin Lettner',
                    'Martin Pihl',
                    'Masoud Kalali',
                    'mat02',
                    'Matej Urbančič',
                    'Mathias-K',
                    'Mathieu Arès',
                    'Mathieu D. (MatToufoutu)',
                    'Mathijs',
                    'Matrik',
                    'Matteo Renzulli',
                    'Matteo Settenvini',
                    'Matthew Gadd',
                    'Matthias Benkard',
                    'Matthias Mailänder',
                    'Mattias Ohlsson',
                    'Mauro de Carvalho',
                    'Max Molchanov',
                    'Me',
                    'MercuryCC',
                    'Mert Bozkurt',
                    'Mert Dirik',
                    'MFX',
                    'mhietar',
                    'mibtha',
                    'Michael Budde',
                    'Michael Kaliszka',
                    'Michalis Makaronides',
                    'Michał Tokarczyk',
                    'Miguel Pires da Rosa',
                    'Mihai Capotă',
                    'Miika Metsälä',
                    'Mikael Fernblad',
                    'Mike Sierra',
                    'mikhalek',
                    'Milan Prvulović',
                    'Milo Casagrande',
                    'Mindaugas',
                    'Miroslav Matejaš',
                    'misel',
                    'mithras',
                    'Mitja Pagon',
                    'M.Kitchen',
                    'Mohamed Magdy',
                    'moonkey',
                    'MrBlonde',
                    'muczy',
                    'Münir Ekinci',
                    'Mustafa Temizel',
                    'mvoncken',
                    'Mytonn',
                    'NagyMarton',
                    'neaion',
                    'Neil Lin',
                    'Nemo',
                    'Nerijus Arlauskas',
                    'Nicklas Larsson',
                    'Nicolaj Wyke',
                    'Nicola Piovesan',
                    'Nicolas Sabatier',
                    'Nicolas Velin',
                    'Nightfall',
                    'NiKoB',
                    'Nikolai M. Riabov',
                    'Niko_Thien',
                    'niska',
                    'Nithir',
                    'noisemonkey',
                    'nomemohes',
                    'nosense',
                    'null',
                    'Nuno Estêvão',
                    'Nuno Santos',
                    'nxxs',
                    'nyo',
                    'obo',
                    'Ojan',
                    'Olav Andreas Lindekleiv',
                    'oldbeggar',
                    'Olivier FAURAX',
                    'orphe',
                    'osantana',
                    'Osman Tosun',
                    'OssiR',
                    'otypoks',
                    'ounn',
                    'Oz123',
                    'Özgür BASKIN',
                    'Pablo Carmona A.',
                    'Pablo Ledesma',
                    'Pablo Navarro Castillo',
                    'Paco Molinero',
                    'Pål-Eivind Johnsen',
                    'pano',
                    'Paolo Naldini',
                    'Paracelsus',
                    'Patryk13_03',
                    'Patryk Skorupa',
                    'PattogoTehen',
                    'Paul Lange',
                    'Pavcio',
                    'Paweł Wysocki',
                    'Pedro Brites Moita',
                    'Pedro Clemente Pereira Neto',
                    'Pekka "PEXI" Niemistö',
                    'Penegal',
                    'Penzo',
                    'perdido',
                    'Peter Kotrcka',
                    'Peter Skov',
                    'Peter Van den Bosch',
                    'Petter Eklund',
                    'Petter Viklund',
                    'phatsphere',
                    'Phenomen',
                    'Philipi',
                    'Philippides Homer',
                    'phoenix',
                    'pidi',
                    'Pierre Quillery',
                    'Pierre Rudloff',
                    'Pierre Slamich',
                    'Pietrao',
                    'Piotr Strębski',
                    'Piotr Wicijowski',
                    'Pittmann Tamás',
                    'Playmolas',
                    'Prescott',
                    'Prescott_SK',
                    'pronull',
                    'Przemysław Kulczycki',
                    'Pumy',
                    'pushpika',
                    'PY',
                    'qubicllj',
                    'r21vo',
                    'Rafał Barański',
                    'rainofchaos',
                    'Rajbir',
                    'ras0ir',
                    'Rat',
                    'rd1381',
                    'Renato',
                    'Rene Hennig',
                    'Rene Pärts',
                    'Ricardo Duarte',
                    'Richard',
                    'Robert Hrovat',
                    'Roberth Sjonøy',
                    'Robert Lundmark',
                    'Robin Jakobsson',
                    'Robin Kåveland',
                    'Rodrigo Donado',
                    'Roel Groeneveld',
                    'rohmaru',
                    'Rolf Christensen',
                    'Rolf Leggewie',
                    'Roni Kantis',
                    'Ronmi',
                    'Rostislav Raykov',
                    'royto',
                    'RuiAmaro',
                    'Rui Araújo',
                    'Rui Moura',
                    'Rune Svendsen',
                    'Rusna',
                    'Rytis',
                    'Sabirov Mikhail',
                    'salseeg',
                    'Sami Koskinen',
                    'Samir van de Sand',
                    'Samuel Arroyo Acuña',
                    'Samuel R. C. Vale',
                    'Sanel',
                    'Santi',
                    'Santi Martínez Cantelli',
                    'Sardan',
                    'Sargate Kanogan',
                    'Sarmad Jari',
                    'Saša Bodiroža',
                    'sat0shi',
                    'Saulius Pranckevičius',
                    'Savvas Radevic',
                    'Sebastian Krauß',
                    'Sebastián Porta',
                    'Sedir',
                    'Sefa Denizoğlu',
                    'sekolands',
                    'Selim Suerkan',
                    'semsomi',
                    'Sergii Golovatiuk',
                    'setarcos',
                    'Sheki',
                    'Shironeko',
                    'Shlomil',
                    'silfiriel',
                    'Simone Tolotti',
                    'Simone Vendemia',
                    'sirkubador',
                    'Sławomir Więch',
                    'slip',
                    'slyon',
                    'smoke',
                    'Sonja',
                    'spectral',
                    'spin_555',
                    'spitf1r3',
                    'Spiziuz',
                    'Spyros Theodoritsis',
                    'SqUe',
                    'Squigly',
                    'srtck',
                    'Stefan Horning',
                    'Stefano Maggiolo',
                    'Stefano Roberto Soleti',
                    'steinberger',
                    'Stéphane Travostino',
                    'Stephan Klein',
                    'Steven De Winter',
                    'Stevie',
                    'Stian24',
                    'stylius',
                    'Sukarn Maini',
                    'Sunjae Park',
                    'Susana Pereira',
                    'szymon siglowy',
                    'takercena',
                    'TAS',
                    'Taygeto',
                    'temy4',
                    'texxxxxx',
                    'thamood',
                    'Thanos Chatziathanassiou',
                    'Tharawut Paripaiboon',
                    'Theodoor',
                    'Théophane Anestis',
                    'Thor Marius K. Høgås',
                    'Tiago Silva',
                    'Tiago Sousa',
                    'Tikkel',
                    'tim__b',
                    'Tim Bordemann',
                    'Tim Fuchs',
                    'Tim Kornhammar',
                    'Timo',
                    'Timo Jyrinki',
                    'Timothy Babych',
                    'TitkosRejtozo',
                    'Tom',
                    'Tomas Gustavsson',
                    'Tomas Valentukevičius',
                    'Tomasz Dominikowski',
                    'Tomislav Plavčić',
                    'Tom Mannerhagen',
                    'Tommy Mikkelsen',
                    'Tom Verdaat',
                    'Tony Manco',
                    'Tor Erling H. Opsahl',
                    'Toudi',
                    'tqm_z',
                    'Trapanator',
                    'Tribaal',
                    'Triton',
                    'TuniX12',
                    'Tuomo Sipola',
                    'turbojugend_gr',
                    'Turtle.net',
                    'twilight',
                    'tymmej',
                    'Ulrik',
                    'Umarzuki Mochlis',
                    'unikob',
                    'Vadim Gusev',
                    'Vagi',
                    'Valentin Bora',
                    'Valmantas Palikša',
                    'VASKITTU',
                    'Vassilis Skoullis',
                    'vetal17',
                    'vicedo',
                    'viki',
                    'villads hamann',
                    'Vincent Garibal',
                    'Vincent Ortalda',
                    'vinchi007',
                    'Vinícius de Figueiredo Silva',
                    'Vinzenz Vietzke',
                    'virtoo',
                    'virtual_spirit',
                    'Vitor Caike',
                    'Vitor Lamas Gatti',
                    'Vladimir Lazic',
                    'Vladimir Sharshov',
                    'Wanderlust',
                    'Wander Nauta',
                    'Ward De Ridder',
                    'WebCrusader',
                    'webdr',
                    'Wentao Tang',
                    'wilana',
                    'Wilfredo Ernesto Guerrero Campos',
                    'Wim Champagne',
                    'World Sucks',
                    'Xabi Ezpeleta',
                    'Xavi de Moner',
                    'XavierToo',
                    'XChesser',
                    'Xiaodong Xu',
                    'xyb',
                    'Yaron',
                    'Yasen Pramatarov',
                    'YesPoX',
                    'Yuren Ju',
                    'Yves MATHIEU',
                    'zekopeko',
                    'zhuqin',
                    'Zissan',
                    'Γιάννης Κατσαμπίρης',
                    'Артём Попов',
                    'Миша',
                    'Шаймарданов Максим',
                    '蔡查理',
                ]
            )
        )
        self.about.set_wrap_license(True)
        self.about.set_license(
            _(
                'This program is free software; you can redistribute it and/or '
                'modify it under the terms of the GNU General Public License as '
                'published by the Free Software Foundation; either version 3 of '
                'the License, or (at your option) any later version. \n\n'
                'This program '
                'is distributed in the hope that it will be useful, but WITHOUT '
                'ANY WARRANTY; without even the implied warranty of '
                'MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU '
                'General Public License for more details. \n\n'
                'You should have received '
                'a copy of the GNU General Public License along with this program; '
                'if not, see <http://www.gnu.org/licenses>. \n\n'
                'In addition, as a '
                'special exception, the copyright holders give permission to link '
                'the code of portions of this program with the OpenSSL library. '
                'You must obey the GNU General Public License in all respects for '
                'all of the code used other than OpenSSL. \n\n'
                'If you modify file(s) '
                'with this exception, you may extend this exception to your '
                'version of the file(s), but you are not obligated to do so. If '
                'you do not wish to do so, delete this exception statement from '
                'your version. If you delete this exception statement from all '
                'source files in the program, then also delete it here.'
            )
        )
        self.about.set_website('http://deluge-torrent.org')
        self.about.set_website_label('deluge-torrent.org')

        self.about.set_icon(get_deluge_icon())
        self.about.set_logo(get_pixbuf('deluge-about.png'))

        if client.connected():
            if not client.is_standalone():
                self.about.set_comments(
                    self.about.get_comments() + _('Server:') + ' %coreversion%\n'
                )

            self.about.set_comments(
                self.about.get_comments() + '\n' + _('libtorrent:') + ' %ltversion%\n'
            )

            def on_lt_version(result):
                c = self.about.get_comments()
                c = c.replace('%ltversion%', result)
                self.about.set_comments(c)

            def on_info(result):
                c = self.about.get_comments()
                c = c.replace('%coreversion%', result)
                self.about.set_comments(c)
                client.core.get_libtorrent_version().addCallback(on_lt_version)

            if not client.is_standalone():
                client.daemon.info().addCallback(on_info)
            else:
                client.core.get_libtorrent_version().addCallback(on_lt_version)

    def run(self):
        self.about.show_all()
        self.about.run()
        self.about.destroy()



#
# Copyright (C) 2007, 2008 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging
import os.path

from gi.repository import Gdk, Gtk
from twisted.internet import defer

import deluge.component as component
from deluge.common import is_url, resource_filename
from deluge.configmanager import ConfigManager
from deluge.ui.client import client

from .common import get_deluge_icon

log = logging.getLogger(__name__)


def last_tier_trackers_from_liststore(trackers_liststore):
    """Create a list of tracker from existing liststore and find last tier number.

    Args:
        tracker_liststore (Gtk.ListStore): A Gtk.ListStore with [tier (int), tracker (str)] rows.

    Returns:
        tuple(int, list): A tuple containing last tier number and list of trackers.

    """

    last_tier = 0
    trackers_from_liststore = []
    for tier, tracker in trackers_liststore:
        trackers_from_liststore.append(tracker)
        if tier >= last_tier:
            last_tier = tier + 1

    return last_tier, trackers_from_liststore


def trackers_tiers_from_text(text_str=''):
    """Create a list of trackers from text.

    Any duplicate trackers are removed.

    Args:
        text_input (str): A block of text with tracker separated by newlines.

    Returns:
        list: The list of trackers.

    Notes:
        Trackers should be separated by newlines and empty line denotes start of new tier.

    """

    trackers = {}
    tier = 0

    lines = text_str.strip().split('\n')
    for line in lines:
        if not line:
            tier += 1
            continue
        line = line.replace('\\', '/')  # Fix any mistyped urls.
        if is_url(line) and line not in trackers:
            trackers[line] = tier

    return trackers


class EditTrackersDialog:
    def __init__(self, torrent_id, parent=None):
        self.torrent_id = torrent_id
        self.builder = Gtk.Builder()
        self.gtkui_config = ConfigManager('gtk3ui.conf')

        # Main dialog
        self.builder.add_from_file(
            resource_filename(__package__, os.path.join('glade', 'edit_trackers.ui'))
        )
        # add tracker dialog
        self.builder.add_from_file(
            resource_filename(
                __package__, os.path.join('glade', 'edit_trackers.add.ui')
            )
        )
        # edit tracker dialog
        self.builder.add_from_file(
            resource_filename(
                __package__, os.path.join('glade', 'edit_trackers.edit.ui')
            )
        )

        self.dialog = self.builder.get_object('edit_trackers_dialog')
        self.treeview = self.builder.get_object('tracker_treeview')
        self.add_tracker_dialog = self.builder.get_object('add_tracker_dialog')
        self.add_tracker_dialog.set_transient_for(self.dialog)
        self.edit_tracker_entry = self.builder.get_object('edit_tracker_entry')
        self.edit_tracker_entry.set_transient_for(self.dialog)
        self.dialog.set_icon(get_deluge_icon())

        self.load_edit_trackers_dialog_state()

        if parent is not None:
            self.dialog.set_transient_for(parent)

        # Connect the signals
        self.builder.connect_signals(self)

        # Create a liststore for tier, url
        self.liststore = Gtk.ListStore(int, str)

        # Create the columns
        self.treeview.append_column(
            Gtk.TreeViewColumn(_('Tier'), Gtk.CellRendererText(), text=0)
        )
        self.treeview.append_column(
            Gtk.TreeViewColumn(_('Tracker'), Gtk.CellRendererText(), text=1)
        )

        self.treeview.set_model(self.liststore)
        self.liststore.set_sort_column_id(0, Gtk.SortType.ASCENDING)

        self.dialog.connect('delete-event', self._on_delete_event)
        self.dialog.connect('response', self._on_response)
        self.treeview.connect('button_press_event', self.on_button_press_event)

        self.add_tracker_dialog.connect('key-press-event', self.on_key_add_press_event)
        self.add_tracker_dialog.connect('delete-event', self.on_delete_event_add)
        self.edit_tracker_entry.connect('key-press-event', self.on_key_edit_press_event)
        self.edit_tracker_entry.connect('delete-event', self.on_delete_event_edit)

    def run(self):
        # Make sure we have a torrent_id.. if not just return
        if self.torrent_id is None:
            return

        # Get the trackers for this torrent
        session = component.get('SessionProxy')
        session.get_torrent_status(self.torrent_id, ['trackers']).addCallback(
            self._on_get_torrent_status
        )
        client.force_call()

        self.deferred = defer.Deferred()
        return self.deferred

    def __del__(self):
        del self.gtkui_config

    def load_edit_trackers_dialog_state(self):
        w = self.gtkui_config['edit_trackers_dialog_width']
        h = self.gtkui_config['edit_trackers_dialog_height']
        if w is not None and h is not None:
            self.dialog.resize(w, h)

    def on_edit_trackers_dialog_configure_event(self, widget, event):
        self.gtkui_config['edit_trackers_dialog_width'] = event.width
        self.gtkui_config['edit_trackers_dialog_height'] = event.height

    def _on_delete_event(self, widget, event):
        self.deferred.callback(Gtk.ResponseType.DELETE_EVENT)
        self.dialog.destroy()

    def _on_response(self, widget, response):
        if response == 1:
            self.trackers = []

            def each(model, path, _iter, data):
                tracker = {}
                tracker['tier'] = model.get_value(_iter, 0)
                tracker['url'] = model.get_value(_iter, 1)
                self.trackers.append(tracker)

            self.liststore.foreach(each, None)
            if self.old_trackers != self.trackers:
                # Set the torrens trackers
                client.core.set_torrent_trackers(self.torrent_id, self.trackers)
                self.deferred.callback(Gtk.ResponseType.OK)
            else:
                self.deferred.callback(Gtk.ResponseType.CANCEL)
        else:
            self.deferred.callback(Gtk.ResponseType.CANCEL)
        self.dialog.destroy()

    def _on_get_torrent_status(self, status):
        """Display trackers dialog"""
        self.old_trackers = list(status['trackers'])
        for tracker in self.old_trackers:
            self.add_tracker(tracker['tier'], tracker['url'])
        self.treeview.set_cursor(0)
        self.dialog.show()

    def add_tracker(self, tier, url):
        """Adds a tracker to the list"""
        self.liststore.append([tier, url])

    def get_selected(self):
        """Returns the selected tracker"""
        return self.treeview.get_selection().get_selected()[1]

    def on_button_add_clicked(self, widget):
        log.debug('on_button_add_clicked')
        # Show the add tracker dialog
        self.add_tracker_dialog.show()
        self.builder.get_object('textview_trackers').grab_focus()
        self.dialog.set_sensitive(False)

    def on_button_remove_clicked(self, widget):
        log.debug('on_button_remove_clicked')
        selected = self.get_selected()
        if selected is not None:
            self.liststore.remove(selected)

    def on_button_edit_clicked(self, widget):
        """edits an existing tracker on edit button click"""
        log.debug('on_button_edit_clicked')
        self._edit_tracker()

    def on_button_press_event(self, widget, event):
        """edits an existing tracker on double click on tracker name"""
        if event.type == Gdk.EventType.DOUBLE_BUTTON_PRESS:
            log.debug('button_press_event double click')
            self._edit_tracker()

    def _edit_tracker(self):
        """edits an existing tracker"""
        selected = self.get_selected()
        if selected:
            tracker = self.liststore.get_value(selected, 1)
            self.builder.get_object('entry_edit_tracker').set_text(tracker)
            self.edit_tracker_entry.show()
            self.edit_tracker_entry.grab_focus()
            self.dialog.set_sensitive(False)

    def _close_edit_dialog(self):
        self.dialog.set_sensitive(True)
        self.edit_tracker_entry.hide()

    def on_button_edit_cancel_clicked(self, widget):
        """handles the cancel button"""
        log.debug('on_button_edit_cancel_clicked')
        self._close_edit_dialog()

    def on_key_edit_press_event(self, widget, event):
        """handles Escape key press"""
        if event.keyval == Gdk.KEY_Escape:
            log.debug('on_key_edit_press_event')
            self._close_edit_dialog()

    def on_delete_event_edit(self, widget, event):
        """handles the Top-Right X button"""
        log.debug('on_delete_event_edit')
        self._close_edit_dialog()
        return True

    def on_button_edit_ok_clicked(self, widget):
        log.debug('on_button_edit_ok_clicked')
        selected = self.get_selected()
        tracker = self.builder.get_object('entry_edit_tracker').get_text()
        self.liststore.set_value(selected, 1, tracker)
        self.dialog.set_sensitive(True)
        self.edit_tracker_entry.hide()

    def on_button_up_clicked(self, widget):
        log.debug('on_button_up_clicked')
        selected = self.get_selected()
        num_rows = self.liststore.iter_n_children(None)
        if selected is not None and num_rows > 1:
            tier = self.liststore.get_value(selected, 0)
            if tier <= 0:
                return
            new_tier = tier - 1
            # Now change the tier for this tracker
            self.liststore.set_value(selected, 0, new_tier)

    def on_button_down_clicked(self, widget):
        log.debug('on_button_down_clicked')
        selected = self.get_selected()
        num_rows = self.liststore.iter_n_children(None)
        if selected is not None and num_rows > 1:
            tier = self.liststore.get_value(selected, 0)
            new_tier = tier + 1
            # Now change the tier for this tracker
            self.liststore.set_value(selected, 0, new_tier)

    def on_button_add_ok_clicked(self, widget):
        log.debug('on_button_add_ok_clicked')

        # Create a list of trackers from the textview widget
        textview_buf = self.builder.get_object('textview_trackers').get_buffer()
        trackers_text = textview_buf.get_text(
            *textview_buf.get_bounds(), include_hidden_chars=False
        )

        for tracker in trackers_tiers_from_text(trackers_text):
            # Figure out what tier number to use.. it's going to be the highest+1
            # Also check for duplicates
            # Check if there are any entries
            duplicate = False
            highest_tier = -1
            for row in self.liststore:
                tier = row[0]
                if tier > highest_tier:
                    highest_tier = tier
                if tracker == row[1]:
                    duplicate = True
                    break

            # If not a duplicate, then add it to the list
            if not duplicate:
                # Add the tracker to the list
                self.add_tracker(highest_tier + 1, tracker)

        # Clear the entry widget and hide the dialog
        textview_buf.set_text('')
        self.dialog.set_sensitive(True)
        self.add_tracker_dialog.hide()

    def _discard_and_close_add_dialog(self):
        # Clear the entry widget and hide the dialog
        b = Gtk.TextBuffer()
        self.builder.get_object('textview_trackers').set_buffer(b)
        self.dialog.set_sensitive(True)
        self.add_tracker_dialog.hide()

    def on_button_add_cancel_clicked(self, widget):
        """handles the cancel button"""
        log.debug('on_button_add_cancel_clicked')
        self._discard_and_close_add_dialog()

    def on_key_add_press_event(self, widget, event):
        """handles Escape key press"""
        if event.keyval == Gdk.KEY_Escape:
            log.debug('on_key_add_press_event')
            self._discard_and_close_add_dialog()

    def on_delete_event_add(self, widget, event):
        """handles the Top-Right X button"""
        log.debug('on_delete_event_add')
        self._discard_and_close_add_dialog()
        return True



#
# Copyright (C) 2007,2008 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import builtins
import ctypes
import gettext
import locale
import logging
import os
import sys
from glob import glob

import deluge.common

from .languages import LANGUAGES

log = logging.getLogger(__name__)

I18N_DOMAIN = 'deluge'


def get_translations_path():
    """Get the absolute path to the directory containing translation files"""
    return deluge.common.resource_filename('deluge', 'i18n')


def get_languages():
    lang = []

    translations_path = get_translations_path()
    lc_messages_path = os.path.join('LC_MESSAGES', I18N_DOMAIN + '.mo')
    for root, dirs, files in os.walk(translations_path):
        # Get the dirs
        lang_dirs = [
            lang_dir
            for lang_dir in dirs
            if glob(os.path.join(translations_path, lang_dir, lc_messages_path))
        ]
        break
    else:
        return lang

    for i, lang_code in enumerate(lang_dirs):
        name = '%s (Language name missing)' % lang_code
        if lang_code in LANGUAGES:
            name = LANGUAGES[lang_code]
        lang.append([lang_code, _(name)])

    lang = sorted(lang, key=lambda l: l[1])
    return lang


def set_language(lang):
    """
    Set the language to use.

    gettext and GtkBuilder will load the translations from the specified
    language.

    :param lang: the language, e.g. "en", "de" or "en_GB"
    :type lang: str
    """
    if not lang:
        return

    # Necessary to set these environment variables for GtkBuilder
    deluge.common.set_env_variable('LANGUAGE', lang)  # Windows/Linux
    deluge.common.set_env_variable('LANG', lang)  # For OSX

    try:
        translation = gettext.translation(
            'deluge', localedir=get_translations_path(), languages=[lang]
        )
    except OSError:
        log.warning('Unable to find translation (.mo) to set language: %s', lang)
    else:
        translation.install()


def setup_mock_translation(warn_msg=None):
    def _func(*txt):
        if warn_msg:
            log.warning(
                '"%s" has been marked for translation, but translation is unavailable.',
                txt[0],
            )
        return txt[0]

    builtins.__dict__['_'] = _func
    builtins.__dict__['ngettext'] = builtins.__dict__['_n'] = _func


# Initialize gettext
def setup_translation():
    translations_path = get_translations_path()
    log.info('Setting up translations from %s', translations_path)

    try:
        if hasattr(locale, 'bindtextdomain'):
            locale.bindtextdomain(I18N_DOMAIN, translations_path)
        if hasattr(locale, 'textdomain'):
            locale.textdomain(I18N_DOMAIN)

        gettext.bindtextdomain(I18N_DOMAIN, translations_path)
        gettext.textdomain(I18N_DOMAIN)

        gettext.install(I18N_DOMAIN, translations_path, names=['ngettext'])
        builtins.__dict__['_n'] = builtins.__dict__['ngettext']

        def load_libintl(libintls):
            errors = []
            libintl = None
            for library in libintls:
                try:
                    libintl = ctypes.cdll.LoadLibrary(library)
                except OSError as ex:
                    errors.append(str(ex))
                else:
                    break

            if not libintl:
                log.debug(
                    'Unable to initialize gettext/locale:\n  %s', '\n  '.join(errors)
                )
                setup_mock_translation()
                return

            return libintl

        libintl = None
        if deluge.common.windows_check():
            libintl = load_libintl(['libintl-8.dll', 'intl.dll'])
        elif deluge.common.osx_check():
            libintl = load_libintl(['libintl.8.dylib', 'libintl.dylib'])

        if libintl:
            libintl.bindtextdomain(
                I18N_DOMAIN, translations_path.encode(sys.getfilesystemencoding())
            )
            libintl.textdomain(I18N_DOMAIN)
            libintl.gettext.restype = ctypes.c_char_p

    except Exception as ex:
        log.error('Unable to initialize gettext/locale!')
        log.exception(ex)
        setup_mock_translation()

    deluge.common.translate_size_units()



#
# Copyright (C) 2007 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#


"""PluginManagerBase"""
import email
import logging
import os.path

import pkg_resources
from twisted.internet import defer
from twisted.python.failure import Failure

import deluge.common
import deluge.component as component
import deluge.configmanager

log = logging.getLogger(__name__)

METADATA_KEYS = [
    'Name',
    'License',
    'Author',
    'Home-page',
    'Summary',
    'Platform',
    'Version',
    'Author-email',
    'Description',
]

DEPRECATION_WARNING = """
The plugin %s is not using the "deluge_" namespace.
In order to avoid package name clashes between regular python packages and
deluge plugins, the way deluge plugins should be created has changed.
If you're seeing this message and you're not the developer of the plugin which
triggered this warning, please report to it's author.
If you're the developer, please take a look at the plugins hosted on deluge's
git repository to have an idea of what needs to be changed.
"""


class PluginManagerBase:
    """PluginManagerBase is a base class for PluginManagers to inherit"""

    def __init__(self, config_file, entry_name):
        log.debug('Plugin manager init..')

        self.config = deluge.configmanager.ConfigManager(config_file)

        # Create the plugins folder if it doesn't exist
        if not os.path.exists(
            os.path.join(deluge.configmanager.get_config_dir(), 'plugins')
        ):
            os.mkdir(os.path.join(deluge.configmanager.get_config_dir(), 'plugins'))

        # This is the entry we want to load..
        self.entry_name = entry_name

        # Loaded plugins
        self.plugins = {}

        # Scan the plugin folders for plugins
        self.scan_for_plugins()

    def enable_plugins(self):
        # Load plugins that are enabled in the config.
        for name in self.config['enabled_plugins']:
            self.enable_plugin(name)

    def disable_plugins(self):
        """Disable all plugins that are enabled"""
        # Dict will be modified so iterate over generated list
        for key in list(self.plugins):
            self.disable_plugin(key)

    def __getitem__(self, key):
        return self.plugins[key]

    def get_available_plugins(self):
        """Returns a list of the available plugins name"""
        return self.available_plugins

    def get_enabled_plugins(self):
        """Returns a list of enabled plugins"""
        return list(self.plugins)

    def scan_for_plugins(self):
        """Scans for available plugins"""
        base_dir = deluge.common.resource_filename('deluge', 'plugins')
        user_dir = os.path.join(deluge.configmanager.get_config_dir(), 'plugins')
        base_subdir = [
            os.path.join(base_dir, f)
            for f in os.listdir(base_dir)
            if os.path.isdir(os.path.join(base_dir, f))
        ]
        plugin_dirs = [base_dir, user_dir] + base_subdir

        for dirname in plugin_dirs:
            pkg_resources.working_set.add_entry(dirname)
        self.pkg_env = pkg_resources.Environment(
            plugin_dirs, platform=None, python=None
        )

        self.available_plugins = []
        for name in self.pkg_env:
            log.debug(
                'Found plugin: %s %s at %s',
                self.pkg_env[name][0].project_name,
                self.pkg_env[name][0].version,
                self.pkg_env[name][0].location,
            )
            self.available_plugins.append(self.pkg_env[name][0].project_name)

    def enable_plugin(self, plugin_name):
        """Enable a plugin.

        Args:
            plugin_name (str): The plugin name.

        Returns:
            Deferred: A deferred with callback value True or False indicating
                whether the plugin is enabled or not.

        """
        if plugin_name not in self.available_plugins:
            log.warning('Cannot enable non-existent plugin %s', plugin_name)
            return defer.succeed(False)

        if plugin_name in self.plugins:
            log.warning('Cannot enable already enabled plugin %s', plugin_name)
            return defer.succeed(True)

        plugin_name = plugin_name.replace(' ', '-')
        egg = self.pkg_env[plugin_name][0]
        # Activate is required by non-namespace plugins.
        egg.activate()
        return_d = defer.succeed(True)

        for name in egg.get_entry_map(self.entry_name):
            try:
                cls = egg.load_entry_point(self.entry_name, name)
                instance = cls(plugin_name.replace('-', '_'))
            except component.ComponentAlreadyRegistered as ex:
                log.error(ex)
                return defer.succeed(False)
            except Exception as ex:
                log.error(
                    'Unable to instantiate plugin %r from %r!', name, egg.location
                )
                log.exception(ex)
                continue
            try:
                return_d = defer.maybeDeferred(instance.enable)
            except Exception as ex:
                log.error('Unable to enable plugin: %s', name)
                log.exception(ex)
                return_d = defer.fail(False)

            if not instance.__module__.startswith('deluge_'):
                import warnings

                warnings.warn_explicit(
                    DEPRECATION_WARNING % name,
                    DeprecationWarning,
                    instance.__module__,
                    0,
                )
            if self._component_state == 'Started':

                def on_enabled(result, instance):
                    return component.start([instance.plugin._component_name])

                return_d.addCallback(on_enabled, instance)

            def on_started(result, instance):
                plugin_name_space = plugin_name.replace('-', ' ')
                self.plugins[plugin_name_space] = instance
                if plugin_name_space not in self.config['enabled_plugins']:
                    log.debug(
                        'Adding %s to enabled_plugins list in config', plugin_name_space
                    )
                    self.config['enabled_plugins'].append(plugin_name_space)
                log.info('Plugin %s enabled...', plugin_name_space)
                return True

            def on_started_error(result, instance):
                log.error(
                    'Failed to start plugin: %s\n%s',
                    plugin_name,
                    result.getTraceback(elideFrameworkCode=1, detail='brief'),
                )
                self.plugins[plugin_name.replace('-', ' ')] = instance
                self.disable_plugin(plugin_name)
                return False

            return_d.addCallbacks(
                on_started,
                on_started_error,
                callbackArgs=[instance],
                errbackArgs=[instance],
            )
            return return_d

        return defer.succeed(False)

    def disable_plugin(self, name):
        """Disable a plugin.

        Args:
            plugin_name (str): The plugin name.

        Returns:
            Deferred: A deferred with callback value True or False indicating
                whether the plugin is disabled or not.

        """
        if name not in self.plugins:
            log.warning('Plugin "%s" is not enabled...', name)
            return defer.succeed(True)

        try:
            d = defer.maybeDeferred(self.plugins[name].disable)
        except Exception as ex:
            log.error('Error when disabling plugin: %s', self.plugin._component_name)
            log.debug(ex)
            d = defer.succeed(False)

        def on_disabled(result):
            ret = True
            if isinstance(result, Failure):
                log.debug(
                    'Error when disabling plugin %s: %s', name, result.getTraceback()
                )
                ret = False
            try:
                component.deregister(self.plugins[name].plugin)
                del self.plugins[name]
                self.config['enabled_plugins'].remove(name)
            except Exception as ex:
                log.warning('Problems occurred disabling plugin: %s', name)
                log.debug(ex)
                ret = False
            else:
                log.info('Plugin %s disabled...', name)
            return ret

        d.addBoth(on_disabled)
        return d

    def get_plugin_info(self, name):
        """Returns a dictionary of plugin info from the metadata"""

        if not self.pkg_env[name]:
            log.warning('Failed to retrieve info for plugin: %s', name)
            info = {}.fromkeys(METADATA_KEYS, '')
            info['Name'] = info['Version'] = 'not available'
            return info

        pkg_info = self.pkg_env[name][0].get_metadata('PKG-INFO')
        return self.parse_pkg_info(pkg_info)

    @staticmethod
    def parse_pkg_info(pkg_info):
        metadata_msg = email.message_from_string(pkg_info)
        metadata_ver = metadata_msg.get('Metadata-Version')

        info = {key: metadata_msg.get(key, '') for key in METADATA_KEYS}

        # Optional Description field in body (Metadata spec >=2.1)
        if not info['Description'] and metadata_ver.startswith('2'):
            info['Description'] = metadata_msg.get_payload().strip()

        return info



from deluge.ui.console.widgets.inputpane import BaseInputPane
from deluge.ui.console.widgets.statusbars import StatusBars
from deluge.ui.console.widgets.window import BaseWindow

__all__ = ['BaseInputPane', 'StatusBars', 'BaseWindow']



#
# Copyright (C) 2007-2009 Andrew Resch <andrewresch@gmail.com>
# Copyright (C) 2011 Pedro Algarvio <pedro@algarvio.me>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import glob
import logging
import os
import shutil
import tempfile
import threading
from base64 import b64decode, b64encode
from typing import Any, Dict, List, Optional, Tuple, Union
from urllib.request import URLError, urlopen

from twisted.internet import defer, reactor, task
from twisted.web.client import Agent, readBody

import deluge.common
import deluge.component as component
from deluge import path_chooser_common
from deluge._libtorrent import LT_VERSION, lt
from deluge.configmanager import ConfigManager, get_config_dir
from deluge.core.alertmanager import AlertManager
from deluge.core.authmanager import (
    AUTH_LEVEL_ADMIN,
    AUTH_LEVEL_NONE,
    AUTH_LEVELS_MAPPING,
    AUTH_LEVELS_MAPPING_REVERSE,
    AuthManager,
)
from deluge.core.eventmanager import EventManager
from deluge.core.filtermanager import FilterManager
from deluge.core.pluginmanager import PluginManager
from deluge.core.preferencesmanager import PreferencesManager
from deluge.core.rpcserver import export
from deluge.core.torrentmanager import TorrentManager
from deluge.decorators import deprecated, maybe_coroutine
from deluge.error import (
    AddTorrentError,
    DelugeError,
    InvalidPathError,
    InvalidTorrentError,
)
from deluge.event import (
    NewVersionAvailableEvent,
    SessionPausedEvent,
    SessionResumedEvent,
    TorrentQueueChangedEvent,
)
from deluge.httpdownloader import download_file

log = logging.getLogger(__name__)

DEPR_SESSION_STATUS_KEYS = {
    # 'active_requests': None, # In dht_stats_alert, if required.
    'allowed_upload_slots': 'ses.num_unchoke_slots',
    # 'dht_global_nodes': None,
    'dht_node_cache': 'dht.dht_node_cache',
    'dht_nodes': 'dht.dht_nodes',
    'dht_torrents': 'dht.dht_torrents',
    # 'dht_total_allocations': None,
    'down_bandwidth_bytes_queue': 'net.limiter_down_bytes',
    'down_bandwidth_queue': 'net.limiter_down_queue',
    'has_incoming_connections': 'net.has_incoming_connections',
    'num_peers': 'peer.num_peers_connected',
    'num_unchoked': 'peer.num_peers_up_unchoked',
    # 'optimistic_unchoke_counter': None, # lt.settings_pack
    'total_dht_download': 'dht.dht_bytes_in',
    'total_dht_upload': 'dht.dht_bytes_out',
    'total_download': 'net.recv_bytes',
    'total_failed_bytes': 'net.recv_failed_bytes',
    'total_ip_overhead_download': 'net.recv_ip_overhead_bytes',
    'total_ip_overhead_upload': 'net.sent_ip_overhead_bytes',
    'total_payload_download': 'net.recv_payload_bytes',
    'total_payload_upload': 'net.sent_payload_bytes',
    'total_redundant_bytes': 'net.recv_redundant_bytes',
    'total_tracker_download': 'net.recv_tracker_bytes',
    'total_tracker_upload': 'net.sent_tracker_bytes',
    'total_upload': 'net.sent_bytes',
    # 'unchoke_counter': None, # lt.settings_pack
    'up_bandwidth_bytes_queue': 'net.limiter_up_bytes',
    'up_bandwidth_queue': 'net.limiter_up_queue',
    # 'utp_stats': None
}

# Session status rate keys associated with session status counters.
SESSION_RATES_MAPPING = {
    'dht_download_rate': 'dht.dht_bytes_in',
    'dht_upload_rate': 'dht.dht_bytes_out',
    'ip_overhead_download_rate': 'net.recv_ip_overhead_bytes',
    'ip_overhead_upload_rate': 'net.sent_ip_overhead_bytes',
    'payload_download_rate': 'net.recv_payload_bytes',
    'payload_upload_rate': 'net.sent_payload_bytes',
    'tracker_download_rate': 'net.recv_tracker_bytes',
    'tracker_upload_rate': 'net.sent_tracker_bytes',
    'download_rate': 'net.recv_bytes',
    'upload_rate': 'net.sent_bytes',
}

DELUGE_VER = deluge.common.get_version()


class Core(component.Component):
    def __init__(
        self, listen_interface=None, outgoing_interface=None, read_only_config_keys=None
    ):
        component.Component.__init__(self, 'Core')

        # Start the libtorrent session.
        user_agent = f'Deluge/{DELUGE_VER} libtorrent/{LT_VERSION}'
        peer_id = self._create_peer_id(DELUGE_VER)
        log.debug('Starting session (peer_id: %s, user_agent: %s)', peer_id, user_agent)
        settings_pack = {
            'peer_fingerprint': peer_id,
            'user_agent': user_agent,
            'ignore_resume_timestamps': True,
        }
        self.session = lt.session(settings_pack, flags=0)

        # Load the settings, if available.
        self._load_session_state()

        # Enable libtorrent extensions
        # Allows peers to download the metadata from the swarm directly
        self.session.add_extension('ut_metadata')
        # Ban peers that sends bad data
        self.session.add_extension('smart_ban')

        # Create the components
        self.eventmanager = EventManager()
        self.preferencesmanager = PreferencesManager()
        self.alertmanager = AlertManager()
        self.pluginmanager = PluginManager(self)
        self.torrentmanager = TorrentManager()
        self.filtermanager = FilterManager(self)
        self.authmanager = AuthManager()

        # New release check information
        self.new_release = None

        # External IP Address from libtorrent
        self.external_ip = None
        self.eventmanager.register_event_handler(
            'ExternalIPEvent', self._on_external_ip_event
        )

        # GeoIP instance with db loaded
        self.geoip_instance = None

        # These keys will be dropped from the set_config() RPC and are
        # configurable from the command-line.
        self.read_only_config_keys = read_only_config_keys
        log.debug('read_only_config_keys: %s', read_only_config_keys)

        # Get the core config
        self.config = ConfigManager('core.conf')
        self.config.save()

        # If there was an interface value from the command line, use it, but
        # store the one in the config so we can restore it on shutdown
        self._old_listen_interface = None
        if listen_interface:
            if deluge.common.is_interface(listen_interface):
                self._old_listen_interface = self.config['listen_interface']
                self.config['listen_interface'] = listen_interface
            else:
                log.error(
                    'Invalid listen interface (must be IP Address or Interface Name): %s',
                    listen_interface,
                )

        self._old_outgoing_interface = None
        if outgoing_interface:
            if deluge.common.is_interface(outgoing_interface):
                self._old_outgoing_interface = self.config['outgoing_interface']
                self.config['outgoing_interface'] = outgoing_interface
            else:
                log.error(
                    'Invalid outgoing interface (must be IP Address or Interface Name): %s',
                    outgoing_interface,
                )

        # New release check information
        self.__new_release = None

        # Session status timer
        self.session_status = {k.name: 0 for k in lt.session_stats_metrics()}
        self._session_prev_bytes = {k: 0 for k in SESSION_RATES_MAPPING}
        # Initiate other session status keys.
        self.session_status.update(self._session_prev_bytes)
        hit_ratio_keys = ['write_hit_ratio', 'read_hit_ratio']
        self.session_status.update({k: 0.0 for k in hit_ratio_keys})

        self.session_status_timer_interval = 0.5
        self.session_status_timer = task.LoopingCall(self.session.post_session_stats)
        self.alertmanager.register_handler(
            'session_stats_alert', self._on_alert_session_stats
        )
        self.session_rates_timer_interval = 2
        self.session_rates_timer = task.LoopingCall(self._update_session_rates)

    def start(self):
        """Starts the core"""
        self.session_status_timer.start(self.session_status_timer_interval)
        self.session_rates_timer.start(self.session_rates_timer_interval, now=False)

    def stop(self):
        log.debug('Core stopping...')

        if self.session_status_timer.running:
            self.session_status_timer.stop()

        if self.session_rates_timer.running:
            self.session_rates_timer.stop()

        # Save the libtorrent session state
        self._save_session_state()

        # We stored a copy of the old interface value
        if self._old_listen_interface is not None:
            self.config['listen_interface'] = self._old_listen_interface

        if self._old_outgoing_interface is not None:
            self.config['outgoing_interface'] = self._old_outgoing_interface

        # Make sure the config file has been saved
        self.config.save()

    def shutdown(self):
        pass

    def apply_session_setting(self, key, value):
        self.apply_session_settings({key: value})

    def apply_session_settings(self, settings):
        """Apply libtorrent session settings.

        Args:
            settings: A dict of lt session settings to apply.
        """
        self.session.apply_settings(settings)

    @staticmethod
    def _create_peer_id(version: str) -> str:
        """Create a peer_id fingerprint.

        This creates the peer_id and modifies the release char to identify
        pre-release and development version. Using ``D`` for dev, daily or
        nightly builds, ``a, b, r`` for pre-releases and ``s`` for
        stable releases.

        Examples:
            ``--<client><client><major><minor><micro><release>--``
            ``--DE200D--`` (development version of 2.0.0)
            ``--DE200s--`` (stable release of v2.0.0)
            ``--DE201b--`` (beta pre-release of v2.0.1)

        Args:
            version: The version string in PEP440 dotted notation.

        Returns:
            The formatted peer_id with Deluge prefix e.g. '--DE200s--'
        """
        split = deluge.common.VersionSplit(version)
        # Fill list with zeros to length of 4 and use lt to create fingerprint.
        version_list = split.version + [0] * (4 - len(split.version))
        peer_id = lt.generate_fingerprint('DE', *version_list)

        def substitute_chr(string, idx, char):
            """Fast substitute single char in string."""
            return string[:idx] + char + string[idx + 1 :]

        if split.dev:
            release_chr = 'D'
        elif split.suffix:
            # a (alpha), b (beta) or r (release candidate).
            release_chr = split.suffix[0].lower()
        else:
            release_chr = 's'
        peer_id = substitute_chr(peer_id, 6, release_chr)

        return peer_id

    def _save_session_state(self):
        """Saves the libtorrent session state"""
        filename = 'session.state'
        filepath = get_config_dir(filename)
        filepath_bak = filepath + '.bak'
        filepath_tmp = filepath + '.tmp'

        try:
            if os.path.isfile(filepath):
                log.debug('Creating backup of %s at: %s', filename, filepath_bak)
                shutil.copy2(filepath, filepath_bak)
        except OSError as ex:
            log.error('Unable to backup %s to %s: %s', filepath, filepath_bak, ex)
        else:
            log.info('Saving the %s at: %s', filename, filepath)
            try:
                with open(filepath_tmp, 'wb') as _file:
                    _file.write(lt.bencode(self.session.save_state()))
                    _file.flush()
                    os.fsync(_file.fileno())
                shutil.move(filepath_tmp, filepath)
            except (OSError, EOFError) as ex:
                log.error('Unable to save %s: %s', filename, ex)
                if os.path.isfile(filepath_bak):
                    log.info('Restoring backup of %s from: %s', filename, filepath_bak)
                    shutil.move(filepath_bak, filepath)

    def _load_session_state(self) -> dict:
        """Loads the libtorrent session state

        Returns:
            A libtorrent sesion state, empty dict if unable to load it.
        """
        filename = 'session.state'
        filepath = get_config_dir(filename)
        filepath_bak = filepath + '.bak'

        for _filepath in (filepath, filepath_bak):
            log.debug('Opening %s for load: %s', filename, _filepath)
            try:
                with open(_filepath, 'rb') as _file:
                    state = lt.bdecode(_file.read())
            except (OSError, EOFError, RuntimeError) as ex:
                log.warning('Unable to load %s: %s', _filepath, ex)
            else:
                log.info('Successfully loaded %s: %s', filename, _filepath)
                self.session.load_state(state)

    def _on_alert_session_stats(self, alert):
        """The handler for libtorrent session stats alert"""
        self.session_status.update(alert.values)
        self._update_session_cache_hit_ratio()

    def _update_session_cache_hit_ratio(self):
        """Calculates the cache read/write hit ratios for session_status."""
        blocks_written = self.session_status['disk.num_blocks_written']
        blocks_read = self.session_status['disk.num_blocks_read']

        if blocks_written:
            self.session_status['write_hit_ratio'] = (
                blocks_written - self.session_status['disk.num_write_ops']
            ) / blocks_written
        else:
            self.session_status['write_hit_ratio'] = 0.0

        if blocks_read:
            self.session_status['read_hit_ratio'] = (
                blocks_read - self.session_status['disk.num_read_ops']
            ) / blocks_read
        else:
            self.session_status['read_hit_ratio'] = 0.0

    def _update_session_rates(self):
        """Calculate session status rates.

        Uses polling interval and counter difference for session_status rates.
        """
        for rate_key, prev_bytes in list(self._session_prev_bytes.items()):
            new_bytes = self.session_status[SESSION_RATES_MAPPING[rate_key]]
            self.session_status[rate_key] = (
                new_bytes - prev_bytes
            ) / self.session_rates_timer_interval
            # Store current value for next update.
            self._session_prev_bytes[rate_key] = new_bytes

    def get_new_release(self):
        log.debug('get_new_release')
        try:
            self.new_release = (
                urlopen('http://download.deluge-torrent.org/version-2.0')
                .read()
                .decode()
                .strip()
            )
        except URLError as ex:
            log.debug('Unable to get release info from website: %s', ex)
        else:
            self.check_new_release()

    def check_new_release(self):
        if self.new_release:
            log.debug('new_release: %s', self.new_release)
            if deluge.common.VersionSplit(
                self.new_release
            ) > deluge.common.VersionSplit(deluge.common.get_version()):
                component.get('EventManager').emit(
                    NewVersionAvailableEvent(self.new_release)
                )
                return self.new_release
        return False

    # Exported Methods
    @export
    def add_torrent_file_async(
        self, filename: str, filedump: str, options: dict, save_state: bool = True
    ) -> 'defer.Deferred[Optional[str]]':
        """Adds a torrent file to the session asynchronously.

        Args:
            filename: The filename of the torrent.
            filedump: A base64 encoded string of torrent file contents.
            options: The options to apply to the torrent upon adding.
            save_state: If the state should be saved after adding the file.

        Returns:
            The torrent ID or None.
        """
        try:
            filedump = b64decode(filedump)
        except TypeError as ex:
            log.error('There was an error decoding the filedump string: %s', ex)

        try:
            d = self.torrentmanager.add_async(
                filedump=filedump,
                options=options,
                filename=filename,
                save_state=save_state,
            )
        except RuntimeError as ex:
            log.error('There was an error adding the torrent file %s: %s', filename, ex)
            raise
        else:
            return d

    @export
    @maybe_coroutine
    async def prefetch_magnet_metadata(
        self, magnet: str, timeout: int = 30
    ) -> Tuple[str, bytes]:
        """Download magnet metadata without adding to Deluge session.

        Used by UIs to get magnet files for selection before adding to session.

        The metadata is bencoded and for transfer base64 encoded.

        Args:
            magnet: The magnet URI.
            timeout: Number of seconds to wait before canceling request.

        Returns:
            A tuple of (torrent_id, metadata) for the magnet.

        """
        return await self.torrentmanager.prefetch_metadata(magnet, timeout)

    @export
    def add_torrent_file(
        self, filename: str, filedump: Union[str, bytes], options: dict
    ) -> Optional[str]:
        """Adds a torrent file to the session.

        Args:
            filename: The filename of the torrent.
            filedump: A base64 encoded string of the torrent file contents.
            options: The options to apply to the torrent upon adding.

        Returns:
            The torrent_id or None.
        """
        try:
            filedump = b64decode(filedump)
        except Exception as ex:
            log.error('There was an error decoding the filedump string: %s', ex)

        try:
            return self.torrentmanager.add(
                filedump=filedump, options=options, filename=filename
            )
        except RuntimeError as ex:
            log.error('There was an error adding the torrent file %s: %s', filename, ex)
            raise

    @export
    def add_torrent_files(
        self, torrent_files: List[Tuple[str, Union[str, bytes], dict]]
    ) -> 'defer.Deferred[List[AddTorrentError]]':
        """Adds multiple torrent files to the session asynchronously.

        Args:
            torrent_files: Torrent files as tuple of
                ``(filename, filedump, options)``.

        Returns:
            A list of errors (if there were any)
        """

        @maybe_coroutine
        async def add_torrents():
            errors = []
            last_index = len(torrent_files) - 1
            for idx, torrent in enumerate(torrent_files):
                try:
                    await self.add_torrent_file_async(
                        torrent[0], torrent[1], torrent[2], save_state=idx == last_index
                    )
                except AddTorrentError as ex:
                    log.warning('Error when adding torrent: %s', ex)
                    errors.append(ex)
            defer.returnValue(errors)

        return task.deferLater(reactor, 0, add_torrents)

    @export
    @maybe_coroutine
    async def add_torrent_url(
        self, url: str, options: dict, headers: dict = None
    ) -> 'defer.Deferred[Optional[str]]':
        """Adds a torrent from a URL. Deluge will attempt to fetch the torrent
        from the URL prior to adding it to the session.

        Args:
            url: the URL pointing to the torrent file
            options: the options to apply to the torrent on add
            headers: any optional headers to send

        Returns:
            a Deferred which returns the torrent_id as a str or None
        """
        log.info('Attempting to add URL %s', url)

        tmp_fd, tmp_file = tempfile.mkstemp(prefix='deluge_url.', suffix='.torrent')
        try:
            filename = await download_file(
                url, tmp_file, headers=headers, force_filename=True
            )
        except Exception:
            log.error('Failed to add torrent from URL %s', url)
            raise
        else:
            with open(filename, 'rb') as _file:
                data = _file.read()
            return self.add_torrent_file(filename, b64encode(data), options)
        finally:
            try:
                os.close(tmp_fd)
                os.remove(tmp_file)
            except OSError as ex:
                log.warning(f'Unable to delete temp file {tmp_file}: , {ex}')

    @export
    def add_torrent_magnet(self, uri: str, options: dict) -> str:
        """Adds a torrent from a magnet link.

        Args:
            uri: the magnet link
            options: the options to apply to the torrent on add

        Returns:
            the torrent_id
        """
        log.debug('Attempting to add by magnet URI: %s', uri)

        return self.torrentmanager.add(magnet=uri, options=options)

    @export
    def remove_torrent(self, torrent_id: str, remove_data: bool) -> bool:
        """Removes a single torrent from the session.

        Args:
            torrent_id: The torrent ID to remove.
            remove_data: If True, also remove the downloaded data.

        Returns:
            True if removed successfully.

        Raises:
             InvalidTorrentError: If the torrent ID does not exist in the session.
        """
        log.debug('Removing torrent %s from the core.', torrent_id)
        return self.torrentmanager.remove(torrent_id, remove_data)

    @export
    def remove_torrents(
        self, torrent_ids: List[str], remove_data: bool
    ) -> 'defer.Deferred[List[Tuple[str, str]]]':
        """Remove multiple torrents from the session.

        Args:
            torrent_ids: The torrent IDs to remove.
            remove_data: If True, also remove the downloaded data.

        Returns:
            An empty list if no errors occurred otherwise the list contains
            tuples of strings, a torrent ID and an error message. For example:

            [('<torrent_id>', 'Error removing torrent')]
        """
        log.info('Removing %d torrents from core.', len(torrent_ids))

        def do_remove_torrents():
            errors = []
            for torrent_id in torrent_ids:
                try:
                    self.torrentmanager.remove(
                        torrent_id, remove_data=remove_data, save_state=False
                    )
                except InvalidTorrentError as ex:
                    errors.append((torrent_id, str(ex)))
            # Save the session state
            self.torrentmanager.save_state()
            if errors:
                log.warning(
                    'Failed to remove %d of %d torrents.', len(errors), len(torrent_ids)
                )
            return errors

        return task.deferLater(reactor, 0, do_remove_torrents)

    @export
    def get_session_status(self, keys: List[str]) -> Dict[str, Union[int, float]]:
        """Gets the session status values for 'keys', these keys are taking
        from libtorrent's session status.

        See: http://www.rasterbar.com/products/libtorrent/manual.html#status

        Args:
            keys: the keys for which we want values

        Returns:
            a dictionary of {key: value, ...}
        """
        if not keys:
            return self.session_status

        status = {}
        for key in keys:
            try:
                status[key] = self.session_status[key]
            except KeyError:
                if key in DEPR_SESSION_STATUS_KEYS:
                    new_key = DEPR_SESSION_STATUS_KEYS[key]
                    log.debug(
                        'Deprecated session status key %s, please use %s', key, new_key
                    )
                    status[key] = self.session_status[new_key]
                else:
                    log.debug('Session status key not valid: %s', key)
        return status

    @export
    def force_reannounce(self, torrent_ids: List[str]) -> None:
        log.debug('Forcing reannouncment to: %s', torrent_ids)
        for torrent_id in torrent_ids:
            self.torrentmanager[torrent_id].force_reannounce()

    @export
    def pause_torrent(self, torrent_id: str) -> None:
        """Pauses a torrent"""
        log.debug('Pausing: %s', torrent_id)
        if not isinstance(torrent_id, str):
            self.pause_torrents(torrent_id)
        else:
            self.torrentmanager[torrent_id].pause()

    @export
    def pause_torrents(self, torrent_ids: List[str] = None) -> None:
        """Pauses a list of torrents"""
        if not torrent_ids:
            torrent_ids = self.torrentmanager.get_torrent_list()
        for torrent_id in torrent_ids:
            self.pause_torrent(torrent_id)

    @export
    def connect_peer(self, torrent_id: str, ip: str, port: int):
        log.debug('adding peer %s to %s', ip, torrent_id)
        if not self.torrentmanager[torrent_id].connect_peer(ip, port):
            log.warning('Error adding peer %s:%s to %s', ip, port, torrent_id)

    @export
    def move_storage(self, torrent_ids: List[str], dest: str):
        log.debug('Moving storage %s to %s', torrent_ids, dest)
        for torrent_id in torrent_ids:
            if not self.torrentmanager[torrent_id].move_storage(dest):
                log.warning('Error moving torrent %s to %s', torrent_id, dest)

    @export
    def pause_session(self) -> None:
        """Pause the entire session"""
        if not self.session.is_paused():
            self.session.pause()
            component.get('EventManager').emit(SessionPausedEvent())

    @export
    def resume_session(self) -> None:
        """Resume the entire session"""
        if self.session.is_paused():
            self.session.resume()
            for torrent_id in self.torrentmanager.torrents:
                self.torrentmanager[torrent_id].update_state()
            component.get('EventManager').emit(SessionResumedEvent())

    @export
    def is_session_paused(self) -> bool:
        """Returns the activity of the session"""
        return self.session.is_paused()

    @export
    def resume_torrent(self, torrent_id: str) -> None:
        """Resumes a torrent"""
        log.debug('Resuming: %s', torrent_id)
        if not isinstance(torrent_id, str):
            self.resume_torrents(torrent_id)
        else:
            self.torrentmanager[torrent_id].resume()

    @export
    def resume_torrents(self, torrent_ids: List[str] = None) -> None:
        """Resumes a list of torrents"""
        if not torrent_ids:
            torrent_ids = self.torrentmanager.get_torrent_list()
        for torrent_id in torrent_ids:
            self.resume_torrent(torrent_id)

    def create_torrent_status(
        self,
        torrent_id,
        torrent_keys,
        plugin_keys,
        diff=False,
        update=False,
        all_keys=False,
    ):
        try:
            status = self.torrentmanager[torrent_id].get_status(
                torrent_keys, diff, update=update, all_keys=all_keys
            )
        except KeyError:
            import traceback

            traceback.print_exc()
            # Torrent was probably removed meanwhile
            return {}

        # Ask the plugin manager to fill in the plugin keys
        if len(plugin_keys) > 0 or all_keys:
            status.update(self.pluginmanager.get_status(torrent_id, plugin_keys))
        return status

    @export
    def get_torrent_status(
        self, torrent_id: str, keys: List[str], diff: bool = False
    ) -> dict:
        torrent_keys, plugin_keys = self.torrentmanager.separate_keys(
            keys, [torrent_id]
        )
        return self.create_torrent_status(
            torrent_id,
            torrent_keys,
            plugin_keys,
            diff=diff,
            update=True,
            all_keys=not keys,
        )

    @export
    @maybe_coroutine
    async def get_torrents_status(
        self, filter_dict: dict, keys: List[str], diff: bool = False
    ) -> dict:
        """returns all torrents , optionally filtered by filter_dict."""
        all_keys = not keys
        torrent_ids = self.filtermanager.filter_torrent_ids(filter_dict)
        status_dict, plugin_keys = await self.torrentmanager.torrents_status_update(
            torrent_ids, keys, diff=diff
        )
        # Ask the plugin manager to fill in the plugin keys
        if len(plugin_keys) > 0 or all_keys:
            for key in status_dict:
                status_dict[key].update(self.pluginmanager.get_status(key, plugin_keys))
        return status_dict

    @export
    def get_filter_tree(
        self, show_zero_hits: bool = True, hide_cat: List[str] = None
    ) -> Dict:
        """returns {field: [(value,count)] }
        for use in sidebar(s)
        """
        return self.filtermanager.get_filter_tree(show_zero_hits, hide_cat)

    @export
    def get_session_state(self) -> List[str]:
        """Returns a list of torrent_ids in the session."""
        # Get the torrent list from the TorrentManager
        return self.torrentmanager.get_torrent_list()

    @export
    def get_config(self) -> dict:
        """Get all the preferences as a dictionary"""
        return self.config.config

    @export
    def get_config_value(self, key: str) -> Any:
        """Get the config value for key"""
        return self.config.get(key)

    @export
    def get_config_values(self, keys: List[str]) -> Dict[str, Any]:
        """Get the config values for the entered keys"""
        return {key: self.config.get(key) for key in keys}

    @export
    def set_config(self, config: Dict[str, Any]):
        """Set the config with values from dictionary"""
        # Load all the values into the configuration
        for key in config:
            if self.read_only_config_keys and key in self.read_only_config_keys:
                continue
            self.config[key] = config[key]

    @export
    def get_listen_port(self) -> int:
        """Returns the active listen port"""
        return self.session.listen_port()

    @export
    def get_proxy(self) -> Dict[str, Any]:
        """Returns the proxy settings

        Returns:
            Proxy settings.

        Notes:
            Proxy type names:
                0: None, 1: Socks4, 2: Socks5, 3: Socks5 w Auth, 4: HTTP, 5: HTTP w Auth, 6: I2P
        """

        settings = self.session.get_settings()
        proxy_type = settings['proxy_type']
        proxy_hostname = (
            settings['i2p_hostname'] if proxy_type == 6 else settings['proxy_hostname']
        )
        proxy_port = settings['i2p_port'] if proxy_type == 6 else settings['proxy_port']
        proxy_dict = {
            'type': proxy_type,
            'hostname': proxy_hostname,
            'username': settings['proxy_username'],
            'password': settings['proxy_password'],
            'port': proxy_port,
            'proxy_hostnames': settings['proxy_hostnames'],
            'proxy_peer_connections': settings['proxy_peer_connections'],
            'proxy_tracker_connections': settings['proxy_tracker_connections'],
        }

        return proxy_dict

    @export
    def get_available_plugins(self) -> List[str]:
        """Returns a list of plugins available in the core"""
        return self.pluginmanager.get_available_plugins()

    @export
    def get_enabled_plugins(self) -> List[str]:
        """Returns a list of enabled plugins in the core"""
        return self.pluginmanager.get_enabled_plugins()

    @export
    def enable_plugin(self, plugin: str) -> 'defer.Deferred[bool]':
        return self.pluginmanager.enable_plugin(plugin)

    @export
    def disable_plugin(self, plugin: str) -> 'defer.Deferred[bool]':
        return self.pluginmanager.disable_plugin(plugin)

    @export
    def force_recheck(self, torrent_ids: List[str]) -> None:
        """Forces a data recheck on torrent_ids"""
        for torrent_id in torrent_ids:
            self.torrentmanager[torrent_id].force_recheck()

    @export
    def set_torrent_options(
        self, torrent_ids: List[str], options: Dict[str, Any]
    ) -> None:
        """Sets the torrent options for torrent_ids

        Args:
            torrent_ids: A list of torrent_ids to set the options for.
            options: A dict of torrent options to set. See
                ``torrent.TorrentOptions`` class for valid keys.
        """
        if 'owner' in options and not self.authmanager.has_account(options['owner']):
            raise DelugeError('Username "%s" is not known.' % options['owner'])

        if isinstance(torrent_ids, str):
            torrent_ids = [torrent_ids]

        for torrent_id in torrent_ids:
            self.torrentmanager[torrent_id].set_options(options)

    @export
    def set_torrent_trackers(
        self, torrent_id: str, trackers: List[Dict[str, Any]]
    ) -> None:
        """Sets a torrents tracker list. trackers will be ``[{"url", "tier"}]``"""
        return self.torrentmanager[torrent_id].set_trackers(trackers)

    @export
    def get_magnet_uri(self, torrent_id: str) -> str:
        return self.torrentmanager[torrent_id].get_magnet_uri()

    @deprecated
    @export
    def set_torrent_max_connections(self, torrent_id, value):
        """Deprecated: Use set_torrent_options with 'max_connections'"""
        self.set_torrent_options([torrent_id], {'max_connections': value})

    @deprecated
    @export
    def set_torrent_max_upload_slots(self, torrent_id, value):
        """Deprecated: Use set_torrent_options with 'max_upload_slots'"""
        self.set_torrent_options([torrent_id], {'max_upload_slots': value})

    @deprecated
    @export
    def set_torrent_max_upload_speed(self, torrent_id, value):
        """Deprecated: Use set_torrent_options with 'max_upload_speed'"""
        self.set_torrent_options([torrent_id], {'max_upload_speed': value})

    @deprecated
    @export
    def set_torrent_max_download_speed(self, torrent_id, value):
        """Deprecated: Use set_torrent_options with 'max_download_speed'"""
        self.set_torrent_options([torrent_id], {'max_download_speed': value})

    @deprecated
    @export
    def set_torrent_file_priorities(self, torrent_id, priorities):
        """Deprecated: Use set_torrent_options with 'file_priorities'"""
        self.set_torrent_options([torrent_id], {'file_priorities': priorities})

    @deprecated
    @export
    def set_torrent_prioritize_first_last(self, torrent_id, value):
        """Deprecated: Use set_torrent_options with 'prioritize_first_last'"""
        self.set_torrent_options([torrent_id], {'prioritize_first_last_pieces': value})

    @deprecated
    @export
    def set_torrent_auto_managed(self, torrent_id, value):
        """Deprecated: Use set_torrent_options with 'auto_managed'"""
        self.set_torrent_options([torrent_id], {'auto_managed': value})

    @deprecated
    @export
    def set_torrent_stop_at_ratio(self, torrent_id, value):
        """Deprecated: Use set_torrent_options with 'stop_at_ratio'"""
        self.set_torrent_options([torrent_id], {'stop_at_ratio': value})

    @deprecated
    @export
    def set_torrent_stop_ratio(self, torrent_id, value):
        """Deprecated: Use set_torrent_options with 'stop_ratio'"""
        self.set_torrent_options([torrent_id], {'stop_ratio': value})

    @deprecated
    @export
    def set_torrent_remove_at_ratio(self, torrent_id, value):
        """Deprecated: Use set_torrent_options with 'remove_at_ratio'"""
        self.set_torrent_options([torrent_id], {'remove_at_ratio': value})

    @deprecated
    @export
    def set_torrent_move_completed(self, torrent_id, value):
        """Deprecated: Use set_torrent_options with 'move_completed'"""
        self.set_torrent_options([torrent_id], {'move_completed': value})

    @deprecated
    @export
    def set_torrent_move_completed_path(self, torrent_id, value):
        """Deprecated: Use set_torrent_options with 'move_completed_path'"""
        self.set_torrent_options([torrent_id], {'move_completed_path': value})

    @export
    def get_path_size(self, path):
        """Returns the size of the file or folder 'path' and -1 if the path is
        inaccessible (non-existent or insufficient privileges)"""
        return deluge.common.get_path_size(path)

    @export
    def create_torrent(
        self,
        path,
        tracker,
        piece_length,
        comment,
        target,
        webseeds,
        private,
        created_by,
        trackers,
        add_to_session,
    ):

        log.debug('creating torrent..')
        threading.Thread(
            target=self._create_torrent_thread,
            args=(
                path,
                tracker,
                piece_length,
                comment,
                target,
                webseeds,
                private,
                created_by,
                trackers,
                add_to_session,
            ),
        ).start()

    def _create_torrent_thread(
        self,
        path,
        tracker,
        piece_length,
        comment,
        target,
        webseeds,
        private,
        created_by,
        trackers,
        add_to_session,
    ):
        from deluge import metafile

        metafile.make_meta_file(
            path,
            tracker,
            piece_length,
            comment=comment,
            target=target,
            webseeds=webseeds,
            private=private,
            created_by=created_by,
            trackers=trackers,
        )
        log.debug('torrent created!')
        if add_to_session:
            options = {}
            options['download_location'] = os.path.split(path)[0]
            with open(target, 'rb') as _file:
                filedump = b64encode(_file.read())
                self.add_torrent_file(os.path.split(target)[1], filedump, options)

    @export
    def upload_plugin(self, filename: str, filedump: Union[str, bytes]) -> None:
        """This method is used to upload new plugins to the daemon.  It is used
        when connecting to the daemon remotely and installing a new plugin on
        the client side. ``plugin_data`` is a ``xmlrpc.Binary`` object of the file data,
        i.e. ``plugin_file.read()``"""

        try:
            filedump = b64decode(filedump)
        except TypeError as ex:
            log.error('There was an error decoding the filedump string!')
            log.exception(ex)
            return

        with open(os.path.join(get_config_dir(), 'plugins', filename), 'wb') as _file:
            _file.write(filedump)
        component.get('CorePluginManager').scan_for_plugins()

    @export
    def rescan_plugins(self) -> None:
        """Re-scans the plugin folders for new plugins"""
        component.get('CorePluginManager').scan_for_plugins()

    @export
    def rename_files(
        self, torrent_id: str, filenames: List[Tuple[int, str]]
    ) -> defer.Deferred:
        """Rename files in ``torrent_id``.  Since this is an asynchronous operation by
        libtorrent, watch for the TorrentFileRenamedEvent to know when the
        files have been renamed.

        Args:
            torrent_id: the torrent_id to rename files
            filenames: a list of index, filename pairs

        Raises:
            InvalidTorrentError: if torrent_id is invalid
        """
        if torrent_id not in self.torrentmanager.torrents:
            raise InvalidTorrentError('torrent_id is not in session')

        def rename():
            self.torrentmanager[torrent_id].rename_files(filenames)

        return task.deferLater(reactor, 0, rename)

    @export
    def rename_folder(
        self, torrent_id: str, folder: str, new_folder: str
    ) -> defer.Deferred:
        """Renames the 'folder' to 'new_folder' in 'torrent_id'.  Watch for the
        TorrentFolderRenamedEvent which is emitted when the folder has been
        renamed successfully.

        Args:
            torrent_id: the torrent to rename folder in
            folder: the folder to rename
            new_folder: the new folder name

        Raises:
            InvalidTorrentError: if the torrent_id is invalid
        """
        if torrent_id not in self.torrentmanager.torrents:
            raise InvalidTorrentError('torrent_id is not in session')

        return self.torrentmanager[torrent_id].rename_folder(folder, new_folder)

    @export
    def queue_top(self, torrent_ids: List[str]) -> None:
        log.debug('Attempting to queue %s to top', torrent_ids)
        # torrent_ids must be sorted in reverse before moving to preserve order
        for torrent_id in sorted(
            torrent_ids, key=self.torrentmanager.get_queue_position, reverse=True
        ):
            try:
                # If the queue method returns True, then we should emit a signal
                if self.torrentmanager.queue_top(torrent_id):
                    component.get('EventManager').emit(TorrentQueueChangedEvent())
            except KeyError:
                log.warning('torrent_id: %s does not exist in the queue', torrent_id)

    @export
    def queue_up(self, torrent_ids: List[str]) -> None:
        log.debug('Attempting to queue %s to up', torrent_ids)
        torrents = (
            (self.torrentmanager.get_queue_position(torrent_id), torrent_id)
            for torrent_id in torrent_ids
        )
        torrent_moved = True
        prev_queue_position = None
        # torrent_ids must be sorted before moving.
        for queue_position, torrent_id in sorted(torrents):
            # Move the torrent if and only if there is space (by not moving it we preserve the order)
            if torrent_moved or queue_position - prev_queue_position > 1:
                try:
                    torrent_moved = self.torrentmanager.queue_up(torrent_id)
                except KeyError:
                    log.warning(
                        'torrent_id: %s does not exist in the queue', torrent_id
                    )
            # If the torrent moved, then we should emit a signal
            if torrent_moved:
                component.get('EventManager').emit(TorrentQueueChangedEvent())
            else:
                prev_queue_position = queue_position

    @export
    def queue_down(self, torrent_ids: List[str]) -> None:
        log.debug('Attempting to queue %s to down', torrent_ids)
        torrents = (
            (self.torrentmanager.get_queue_position(torrent_id), torrent_id)
            for torrent_id in torrent_ids
        )
        torrent_moved = True
        prev_queue_position = None
        # torrent_ids must be sorted before moving.
        for queue_position, torrent_id in sorted(torrents, reverse=True):
            # Move the torrent if and only if there is space (by not moving it we preserve the order)
            if torrent_moved or prev_queue_position - queue_position > 1:
                try:
                    torrent_moved = self.torrentmanager.queue_down(torrent_id)
                except KeyError:
                    log.warning(
                        'torrent_id: %s does not exist in the queue', torrent_id
                    )
            # If the torrent moved, then we should emit a signal
            if torrent_moved:
                component.get('EventManager').emit(TorrentQueueChangedEvent())
            else:
                prev_queue_position = queue_position

    @export
    def queue_bottom(self, torrent_ids: List[str]) -> None:
        log.debug('Attempting to queue %s to bottom', torrent_ids)
        # torrent_ids must be sorted before moving to preserve order
        for torrent_id in sorted(
            torrent_ids, key=self.torrentmanager.get_queue_position
        ):
            try:
                # If the queue method returns True, then we should emit a signal
                if self.torrentmanager.queue_bottom(torrent_id):
                    component.get('EventManager').emit(TorrentQueueChangedEvent())
            except KeyError:
                log.warning('torrent_id: %s does not exist in the queue', torrent_id)

    @export
    def glob(self, path: str) -> List[str]:
        return glob.glob(path)

    @export
    def test_listen_port(self) -> 'defer.Deferred[Optional[bool]]':
        """Checks if the active port is open

        Returns:
            True if the port is open, False if not
        """
        port = self.get_listen_port()
        url = 'https://deluge-torrent.org/test_port.php?port=%s' % port
        agent = Agent(reactor, connectTimeout=30)
        d = agent.request(b'GET', url.encode())

        def on_get_page(body):
            return bool(int(body))

        def on_error(failure):
            log.warning('Error testing listen port: %s', failure)

        d.addCallback(readBody).addCallback(on_get_page)
        d.addErrback(on_error)

        return d

    @export
    def get_free_space(self, path: str = None) -> int:
        """Returns the number of free bytes at path

        Args:
            path: the path to check free space at, if None, use the default download location

        Returns:
            the number of free bytes at path

        Raises:
            InvalidPathError: if the path is invalid
        """
        if not path:
            path = self.config['download_location']
        try:
            return deluge.common.free_space(path)
        except InvalidPathError:
            return -1

    def _on_external_ip_event(self, external_ip):
        self.external_ip = external_ip

    @export
    def get_external_ip(self) -> str:
        """Returns the external IP address received from libtorrent."""
        return self.external_ip

    @export
    def get_libtorrent_version(self) -> str:
        """Returns the libtorrent version.

        Returns:
            the version
        """
        return LT_VERSION

    @export
    def get_completion_paths(self, args: Dict[str, Any]) -> Dict[str, Any]:
        """Returns the available path completions for the input value."""
        return path_chooser_common.get_completion_paths(args)

    @export(AUTH_LEVEL_ADMIN)
    def get_known_accounts(self) -> List[Dict[str, Any]]:
        return self.authmanager.get_known_accounts()

    @export(AUTH_LEVEL_NONE)
    def get_auth_levels_mappings(self) -> Tuple[Dict[str, int], Dict[int, str]]:
        return (AUTH_LEVELS_MAPPING, AUTH_LEVELS_MAPPING_REVERSE)

    @export(AUTH_LEVEL_ADMIN)
    def create_account(self, username: str, password: str, authlevel: str) -> bool:
        return self.authmanager.create_account(username, password, authlevel)

    @export(AUTH_LEVEL_ADMIN)
    def update_account(self, username: str, password: str, authlevel: str) -> bool:
        return self.authmanager.update_account(username, password, authlevel)

    @export(AUTH_LEVEL_ADMIN)
    def remove_account(self, username: str) -> bool:
        return self.authmanager.remove_account(username)



#
# Copyright (C) 2009 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging
import re

from deluge.ui.console.utils import format_utils

try:
    import curses
except ImportError:
    pass

log = logging.getLogger(__name__)

colors = [
    'COLOR_BLACK',
    'COLOR_BLUE',
    'COLOR_CYAN',
    'COLOR_GREEN',
    'COLOR_MAGENTA',
    'COLOR_RED',
    'COLOR_WHITE',
    'COLOR_YELLOW',
]

# {(fg, bg): pair_number, ...}
color_pairs = {('white', 'black'): 0}  # Special case, can't be changed

# Some default color schemes
schemes = {
    'input': ('white', 'black'),
    'normal': ('white', 'black'),
    'status': ('yellow', 'blue', 'bold'),
    'info': ('white', 'black', 'bold'),
    'error': ('red', 'black', 'bold'),
    'success': ('green', 'black', 'bold'),
    'event': ('magenta', 'black', 'bold'),
    'selected': ('black', 'white', 'bold'),
    'marked': ('white', 'blue', 'bold'),
    'selectedmarked': ('blue', 'white', 'bold'),
    'header': ('green', 'black', 'bold'),
    'filterstatus': ('green', 'blue', 'bold'),
}

# Colors for various torrent states
state_color = {
    'Seeding': '{!blue,black,bold!}',
    'Downloading': '{!green,black,bold!}',
    'Paused': '{!white,black!}',
    'Checking': '{!green,black!}',
    'Queued': '{!yellow,black!}',
    'Error': '{!red,black,bold!}',
    'Moving': '{!green,black,bold!}',
}

type_color = {
    bool: '{!yellow,black,bold!}',
    int: '{!green,black,bold!}',
    float: '{!green,black,bold!}',
    str: '{!cyan,black,bold!}',
    list: '{!magenta,black,bold!}',
    dict: '{!white,black,bold!}',
}

tab_char = '\t'
color_tag_start = '{!'
color_tag_end = '!}'


def get_color_pair(fg, bg):
    return color_pairs[(fg, bg)]


def init_colors():
    curses.start_color()

    # We want to redefine white/black as it makes underlining work for some terminals
    # but can also fail on others, so we try/except

    def define_pair(counter, fg_name, bg_name, fg, bg):
        try:
            curses.init_pair(counter, fg, bg)
            color_pairs[(fg_name, bg_name)] = counter
            counter += 1
        except (curses.error, ValueError) as ex:
            log.debug(f'Color pair {fg_name} {bg_name} not available: {ex}')
        return counter

    # Create the color_pairs dict
    counter = 1
    for fg in colors:
        for bg in colors:
            counter = define_pair(
                counter,
                fg[6:].lower(),
                bg[6:].lower(),
                getattr(curses, fg),
                getattr(curses, bg),
            )

    counter = define_pair(counter, 'white', 'grey', curses.COLOR_WHITE, 241)
    counter = define_pair(counter, 'black', 'whitegrey', curses.COLOR_BLACK, 249)
    counter = define_pair(counter, 'magentadark', 'white', 99, curses.COLOR_WHITE)


class BadColorString(Exception):
    pass


def check_tag_count(string):
    """Raise BadColorString if color tag open/close not equal."""
    if string.count(color_tag_start) != string.count(color_tag_end):
        raise BadColorString('Number of {! is not equal to number of !}')


def replace_tabs(line):
    """
    Returns a string with tabs replaced with spaces.

    """
    for i in range(line.count(tab_char)):
        tab_length = 8 - (len(line[: line.find(tab_char)]) % 8)
        line = line.replace(tab_char, b' ' * tab_length, 1)
    return line


def strip_colors(line):
    """
    Returns a string with the color formatting removed.

    """
    check_tag_count(line)

    # Remove all the color tags
    while line.find(color_tag_start) != -1:
        tag_start = line.find(color_tag_start)
        tag_end = line.find(color_tag_end) + 2
        line = line[:tag_start] + line[tag_end:]

    return line


def get_line_length(line):
    """
    Returns the string length without the color formatting.

    """
    # Remove all the color tags
    line = strip_colors(line)

    # Replace tabs with the appropriate amount of spaces
    line = replace_tabs(line)
    return len(line)


def get_line_width(line):
    """
    Get width of string considering double width characters

    """
    # Remove all the color tags
    line = strip_colors(line)

    # Replace tabs with the appropriate amount of spaces
    line = replace_tabs(line)
    return format_utils.strwidth(line)


def parse_color_string(string):
    """Parses a string and returns a list of 2-tuples (color, string).

    Args:
        string (str): The string to parse.
    """
    check_tag_count(string)

    ret = []
    last_color_attr = None
    # Keep track of where the strings
    while string.find(color_tag_start) != -1:
        begin = string.find(color_tag_start)
        if begin > 0:
            ret.append(
                (
                    curses.color_pair(
                        color_pairs[(schemes['input'][0], schemes['input'][1])]
                    ),
                    string[:begin],
                )
            )

        end = string.find(color_tag_end)
        if end == -1:
            raise BadColorString('Missing closing "!}"')

        # Get a list of attributes in the bracketed section
        attrs = string[begin + 2 : end].split(',')

        if len(attrs) == 1 and not attrs[0].strip(' '):
            raise BadColorString('No description in {! !}')

        def apply_attrs(cp, attrs):
            # This function applies any additional attributes as necessary
            for attr in attrs:
                if attr == 'ignore':
                    continue
                mode = '+'
                if attr[0] in ['+', '-']:
                    mode = attr[0]
                    attr = attr[1:]
                if mode == '+':
                    cp |= getattr(curses, 'A_' + attr.upper())
                else:
                    cp ^= getattr(curses, 'A_' + attr.upper())
            return cp

        # Check for a builtin type first
        if attrs[0] in schemes:
            pair = (schemes[attrs[0]][0], schemes[attrs[0]][1])
            if pair not in color_pairs:
                log.debug('Color pair does not exist: %s, attrs: %s', pair, attrs)
                pair = ('white', 'black')
            # Get the color pair number
            color_pair = curses.color_pair(color_pairs[pair])
            color_pair = apply_attrs(color_pair, schemes[attrs[0]][2:])
            last_color_attr = color_pair
        else:
            attrlist = ['blink', 'bold', 'dim', 'reverse', 'standout', 'underline']

            if attrs[0][0] in ['+', '-']:
                # Color is not given, so use last color
                if last_color_attr is None:
                    raise BadColorString(
                        'No color value given when no previous color was used!: %s'
                        % (attrs[0])
                    )
                color_pair = last_color_attr
                for i, attr in enumerate(attrs):
                    if attr[1:] not in attrlist:
                        raise BadColorString('Bad attribute value!: %s' % (attr))
            else:
                # This is a custom color scheme
                fg = attrs[0]
                bg = 'black'  # Default to 'black' if no bg is chosen
                if len(attrs) > 1:
                    bg = attrs[1]
                try:
                    pair = (fg, bg)
                    if pair not in color_pairs:
                        # Color pair missing, this could be because the
                        # terminal settings allows no colors. If background is white, we
                        # assume this means selection, and use "white", "black" + reverse
                        # To have white background and black foreground
                        log.debug('Color pair does not exist: %s', pair)
                        if pair[1] == 'white':
                            if attrs[2] == 'ignore':
                                attrs[2] = 'reverse'
                            else:
                                attrs.append('reverse')
                        pair = ('white', 'black')
                    color_pair = curses.color_pair(color_pairs[pair])
                    last_color_attr = color_pair
                    attrs = attrs[2:]  # Remove colors
                except KeyError:
                    raise BadColorString(f'Bad color value in tag: {fg},{bg}')
            # Check for additional attributes and OR them to the color_pair
            color_pair = apply_attrs(color_pair, attrs)
            last_color_attr = color_pair
        # We need to find the text now, so lets try to find another {! and if
        # there isn't one, then it's the rest of the string
        next_begin = string.find(color_tag_start, end)

        if next_begin == -1:
            ret.append((color_pair, replace_tabs(string[end + 2 :])))
            break
        else:
            ret.append((color_pair, replace_tabs(string[end + 2 : next_begin])))
            string = string[next_begin:]

    if not ret:
        # There was no color scheme so we add it with a 0 for white on black
        ret = [(0, string)]
    return ret


class ConsoleColorFormatter:
    """
    Format help in a way suited to deluge CmdLine mode - colors, format, indentation...
    """

    replace_dict = {
        '<torrent-id>': '{!green!}%s{!input!}',
        '<torrent>': '{!green!}%s{!input!}',
        '<command>': '{!green!}%s{!input!}',
        '<state>': '{!yellow!}%s{!input!}',
        '\\.\\.\\.': '{!yellow!}%s{!input!}',
        '\\s\\*\\s': '{!blue!}%s{!input!}',
        '(?<![\\-a-z])(-[a-zA-Z0-9])': '{!red!}%s{!input!}',
        # "(\-[a-zA-Z0-9])": "{!red!}%s{!input!}",
        '--[_\\-a-zA-Z0-9]+': '{!green!}%s{!input!}',
        '(\\[|\\])': '{!info!}%s{!input!}',
        '<tab>': '{!white!}%s{!input!}',
        '[_A-Z]{3,}': '{!cyan!}%s{!input!}',
        '<key>': '{!cyan!}%s{!input!}',
        '<value>': '{!cyan!}%s{!input!}',
        'usage:': '{!info!}%s{!input!}',
        '<download-folder>': '{!yellow!}%s{!input!}',
        '<torrent-file>': '{!green!}%s{!input!}',
    }

    def format_colors(self, string):
        def r(repl):
            return lambda s: repl % s.group()

        for key, replacement in self.replace_dict.items():
            string = re.sub(key, r(replacement), string)
        return string



#
# Copyright (C) 2008-2010 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#


import logging
import os
import platform
import random
import threading
from urllib.parse import quote_plus
from urllib.request import urlopen

from twisted.internet.task import LoopingCall

import deluge.common
import deluge.component as component
import deluge.configmanager
from deluge._libtorrent import lt
from deluge.event import ConfigValueChangedEvent

GeoIP = None
try:
    from GeoIP import GeoIP
except ImportError:
    try:
        from pygeoip import GeoIP
    except ImportError:
        pass

log = logging.getLogger(__name__)

DEFAULT_PREFS = {
    'send_info': False,
    'info_sent': 0.0,
    'daemon_port': 58846,
    'allow_remote': False,
    'pre_allocate_storage': False,
    'download_location': deluge.common.get_default_download_dir(),
    'listen_ports': [6881, 6891],
    'listen_interface': '',
    'outgoing_interface': '',
    'random_port': True,
    'listen_random_port': None,
    'listen_use_sys_port': False,
    'listen_reuse_port': True,
    'outgoing_ports': [0, 0],
    'random_outgoing_ports': True,
    'copy_torrent_file': False,
    'del_copy_torrent_file': False,
    'torrentfiles_location': deluge.common.get_default_download_dir(),
    'plugins_location': os.path.join(deluge.configmanager.get_config_dir(), 'plugins'),
    'prioritize_first_last_pieces': False,
    'sequential_download': False,
    'dht': True,
    'upnp': True,
    'natpmp': True,
    'utpex': True,
    'lsd': True,
    'enc_in_policy': 1,
    'enc_out_policy': 1,
    'enc_level': 2,
    'max_connections_global': 200,
    'max_upload_speed': -1.0,
    'max_download_speed': -1.0,
    'max_upload_slots_global': 4,
    'max_half_open_connections': (
        lambda: deluge.common.windows_check()
        and (lambda: deluge.common.vista_check() and 4 or 8)()
        or 50
    )(),
    'max_connections_per_second': 20,
    'ignore_limits_on_local_network': True,
    'max_connections_per_torrent': -1,
    'max_upload_slots_per_torrent': -1,
    'max_upload_speed_per_torrent': -1,
    'max_download_speed_per_torrent': -1,
    'enabled_plugins': [],
    'add_paused': False,
    'max_active_seeding': 5,
    'max_active_downloading': 3,
    'max_active_limit': 8,
    'dont_count_slow_torrents': False,
    'queue_new_to_top': False,
    'stop_seed_at_ratio': False,
    'remove_seed_at_ratio': False,
    'stop_seed_ratio': 2.00,
    'share_ratio_limit': 2.00,
    'seed_time_ratio_limit': 7.00,
    'seed_time_limit': 180,
    'auto_managed': True,
    'move_completed': False,
    'move_completed_path': deluge.common.get_default_download_dir(),
    'move_completed_paths_list': [],
    'download_location_paths_list': [],
    'path_chooser_show_chooser_button_on_localhost': True,
    'path_chooser_auto_complete_enabled': True,
    'path_chooser_accelerator_string': 'Tab',
    'path_chooser_max_popup_rows': 20,
    'path_chooser_show_hidden_files': False,
    'new_release_check': True,
    'proxy': {
        'type': 0,
        'hostname': '',
        'username': '',
        'password': '',
        'port': 8080,
        'proxy_hostnames': True,
        'proxy_peer_connections': True,
        'proxy_tracker_connections': True,
        'force_proxy': False,
        'anonymous_mode': False,
    },
    'peer_tos': '0x00',
    'rate_limit_ip_overhead': True,
    'geoip_db_location': '/usr/share/GeoIP/GeoIP.dat',
    'cache_size': 512,
    'cache_expiry': 60,
    'auto_manage_prefer_seeds': False,
    'shared': False,
    'super_seeding': False,
}


class PreferencesManager(component.Component):
    def __init__(self):
        component.Component.__init__(self, 'PreferencesManager')
        self.config = deluge.configmanager.ConfigManager('core.conf', DEFAULT_PREFS)
        if 'proxies' in self.config:
            log.warning(
                'Updating config file for proxy, using "peer" values to fill new "proxy" setting'
            )
            self.config['proxy'].update(self.config['proxies']['peer'])
            log.warning('New proxy config is: %s', self.config['proxy'])
            del self.config['proxies']
        if 'i2p_proxy' in self.config and self.config['i2p_proxy']['hostname']:
            self.config['proxy'].update(self.config['i2p_proxy'])
            self.config['proxy']['type'] = 6
            del self.config['i2p_proxy']
        if 'anonymous_mode' in self.config:
            self.config['proxy']['anonymous_mode'] = self.config['anonymous_mode']
            del self.config['anonymous_mode']
        if 'proxy' in self.config:
            for key in DEFAULT_PREFS['proxy']:
                if key not in self.config['proxy']:
                    self.config['proxy'][key] = DEFAULT_PREFS['proxy'][key]

        self.core = component.get('Core')
        self.new_release_timer = None

    def start(self):
        # Set the initial preferences on start-up
        for key in DEFAULT_PREFS:
            self.do_config_set_func(key, self.config[key])

        self.config.register_change_callback(self._on_config_value_change)

    def stop(self):
        if self.new_release_timer and self.new_release_timer.running:
            self.new_release_timer.stop()

    # Config set functions
    def do_config_set_func(self, key, value):
        on_set_func = getattr(self, '_on_set_' + key, None)
        if on_set_func:
            if log.isEnabledFor(logging.DEBUG):
                log.debug('Config key: %s set to %s..', key, value)
            on_set_func(key, value)

    def _on_config_value_change(self, key, value):
        if self.get_state() == 'Started':
            self.do_config_set_func(key, value)
            component.get('EventManager').emit(ConfigValueChangedEvent(key, value))

    def _on_set_torrentfiles_location(self, key, value):
        if self.config['copy_torrent_file']:
            try:
                os.makedirs(value)
            except OSError as ex:
                log.debug('Unable to make directory: %s', ex)

    def _on_set_listen_ports(self, key, value):
        self.__set_listen_on()

    def _on_set_listen_interface(self, key, value):
        self.__set_listen_on()

    def _on_set_outgoing_interface(self, key, value):
        """Set interface name or IP address for outgoing BitTorrent connections."""
        value = value.strip() if value else ''
        self.core.apply_session_settings({'outgoing_interfaces': value})

    def _on_set_random_port(self, key, value):
        self.__set_listen_on()

    def __set_listen_on(self):
        """Set the ports and interface address to listen for incoming connections on."""
        if self.config['random_port']:
            if not self.config['listen_random_port']:
                self.config['listen_random_port'] = random.randrange(49152, 65525)
            listen_ports = [
                self.config['listen_random_port']
            ] * 2  # use single port range
        else:
            self.config['listen_random_port'] = None
            listen_ports = self.config['listen_ports']

        if self.config['listen_interface']:
            interface = self.config['listen_interface'].strip()
        else:
            interface = '0.0.0.0'

        log.debug(
            'Listen Interface: %s, Ports: %s with use_sys_port: %s',
            interface,
            listen_ports,
            self.config['listen_use_sys_port'],
        )
        interfaces = [
            f'{interface}:{port}'
            for port in range(listen_ports[0], listen_ports[1] + 1)
        ]
        self.core.apply_session_settings(
            {
                'listen_system_port_fallback': self.config['listen_use_sys_port'],
                'listen_interfaces': ','.join(interfaces),
            }
        )

    def _on_set_outgoing_ports(self, key, value):
        self.__set_outgoing_ports()

    def _on_set_random_outgoing_ports(self, key, value):
        self.__set_outgoing_ports()

    def __set_outgoing_ports(self):
        port = (
            0
            if self.config['random_outgoing_ports']
            else self.config['outgoing_ports'][0]
        )
        if port:
            num_ports = (
                self.config['outgoing_ports'][1] - self.config['outgoing_ports'][0]
            )
            num_ports = num_ports if num_ports > 1 else 5
        else:
            num_ports = 0
        log.debug('Outgoing port set to %s with range: %s', port, num_ports)
        self.core.apply_session_settings(
            {'outgoing_port': port, 'num_outgoing_ports': num_ports}
        )

    def _on_set_peer_tos(self, key, value):
        try:
            self.core.apply_session_setting('peer_tos', int(value, 16))
        except ValueError as ex:
            log.error('Invalid tos byte: %s', ex)

    def _on_set_dht(self, key, value):
        lt_bootstraps = self.core.session.get_settings()['dht_bootstrap_nodes']
        # Update list of lt bootstraps, using set to remove duplicates.
        dht_bootstraps = set(
            lt_bootstraps.split(',')
            + [
                'router.bittorrent.com:6881',
                'router.utorrent.com:6881',
                'router.bitcomet.com:6881',
                'dht.transmissionbt.com:6881',
                'dht.aelitis.com:6881',
            ]
        )
        self.core.apply_session_settings(
            {'dht_bootstrap_nodes': ','.join(dht_bootstraps), 'enable_dht': value}
        )

    def _on_set_upnp(self, key, value):
        self.core.apply_session_setting('enable_upnp', value)

    def _on_set_natpmp(self, key, value):
        self.core.apply_session_setting('enable_natpmp', value)

    def _on_set_lsd(self, key, value):
        self.core.apply_session_setting('enable_lsd', value)

    def _on_set_utpex(self, key, value):
        if value:
            self.core.session.add_extension('ut_pex')

    def _on_set_enc_in_policy(self, key, value):
        self._on_set_encryption(key, value)

    def _on_set_enc_out_policy(self, key, value):
        self._on_set_encryption(key, value)

    def _on_set_enc_level(self, key, value):
        self._on_set_encryption(key, value)

    def _on_set_encryption(self, key, value):
        # Convert Deluge enc_level values to libtorrent enc_level values.
        pe_enc_level = {
            0: lt.enc_level.plaintext,
            1: lt.enc_level.rc4,
            2: lt.enc_level.both,
        }
        self.core.apply_session_settings(
            {
                'out_enc_policy': lt.enc_policy(self.config['enc_out_policy']),
                'in_enc_policy': lt.enc_policy(self.config['enc_in_policy']),
                'allowed_enc_level': lt.enc_level(
                    pe_enc_level[self.config['enc_level']]
                ),
                'prefer_rc4': True,
            }
        )

    def _on_set_max_connections_global(self, key, value):
        self.core.apply_session_setting('connections_limit', value)

    def _on_set_max_upload_speed(self, key, value):
        # We need to convert Kb/s to B/s
        value = -1 if value < 0 else int(value * 1024)
        self.core.apply_session_setting('upload_rate_limit', value)

    def _on_set_max_download_speed(self, key, value):
        # We need to convert Kb/s to B/s
        value = -1 if value < 0 else int(value * 1024)
        self.core.apply_session_setting('download_rate_limit', value)

    def _on_set_max_upload_slots_global(self, key, value):
        self.core.apply_session_setting('unchoke_slots_limit', value)

    def _on_set_max_half_open_connections(self, key, value):
        self.core.apply_session_setting('half_open_limit', value)

    def _on_set_max_connections_per_second(self, key, value):
        self.core.apply_session_setting('connection_speed', value)

    def _on_set_ignore_limits_on_local_network(self, key, value):
        self.core.apply_session_setting('ignore_limits_on_local_network', value)

    def _on_set_share_ratio_limit(self, key, value):
        # This value is a float percentage in deluge, but libtorrent needs int percentage.
        self.core.apply_session_setting('share_ratio_limit', int(value * 100))

    def _on_set_seed_time_ratio_limit(self, key, value):
        # This value is a float percentage in deluge, but libtorrent needs int percentage.
        self.core.apply_session_setting('seed_time_ratio_limit', int(value * 100))

    def _on_set_seed_time_limit(self, key, value):
        # This value is stored in minutes in deluge, but libtorrent wants seconds
        self.core.apply_session_setting('seed_time_limit', int(value * 60))

    def _on_set_max_active_downloading(self, key, value):
        self.core.apply_session_setting('active_downloads', value)

    def _on_set_max_active_seeding(self, key, value):
        self.core.apply_session_setting('active_seeds', value)

    def _on_set_max_active_limit(self, key, value):
        self.core.apply_session_setting('active_limit', value)

    def _on_set_dont_count_slow_torrents(self, key, value):
        self.core.apply_session_setting('dont_count_slow_torrents', value)

    def _on_set_send_info(self, key, value):
        """sends anonymous stats home"""
        log.debug('Sending anonymous stats..')

        class SendInfoThread(threading.Thread):
            def __init__(self, config):
                self.config = config
                threading.Thread.__init__(self)

            def run(self):
                import time

                now = time.time()
                # check if we've done this within the last week or never
                if (now - self.config['info_sent']) >= (60 * 60 * 24 * 7):
                    try:
                        url = (
                            'http://deluge-torrent.org/stats_get.php?processor='
                            + platform.machine()
                            + '&python='
                            + platform.python_version()
                            + '&deluge='
                            + deluge.common.get_version()
                            + '&os='
                            + platform.system()
                            + '&plugins='
                            + quote_plus(':'.join(self.config['enabled_plugins']))
                        )
                        urlopen(url)
                    except OSError as ex:
                        log.debug('Network error while trying to send info: %s', ex)
                    else:
                        self.config['info_sent'] = now

        if value:
            SendInfoThread(self.config).start()

    def _on_set_new_release_check(self, key, value):
        if value:
            log.debug('Checking for new release..')
            threading.Thread(target=self.core.get_new_release).start()
            if self.new_release_timer and self.new_release_timer.running:
                self.new_release_timer.stop()
            # Set a timer to check for a new release every 3 days
            self.new_release_timer = LoopingCall(
                self._on_set_new_release_check, 'new_release_check', True
            )
            self.new_release_timer.start(72 * 60 * 60, False)
        else:
            if self.new_release_timer and self.new_release_timer.running:
                self.new_release_timer.stop()

    def _on_set_proxy(self, key, value):
        # Initialise with type none and blank hostnames.
        proxy_settings = {
            'proxy_type': lt.proxy_type_t.none,
            'i2p_hostname': '',
            'proxy_hostname': '',
            'proxy_hostnames': value['proxy_hostnames'],
            'proxy_peer_connections': value['proxy_peer_connections'],
            'proxy_tracker_connections': value['proxy_tracker_connections'],
            'force_proxy': value['force_proxy'],
            'anonymous_mode': value['anonymous_mode'],
        }

        if value['type'] == lt.proxy_type_t.i2p_proxy:
            proxy_settings.update(
                {
                    'proxy_type': lt.proxy_type_t.i2p_proxy,
                    'i2p_hostname': value['hostname'],
                    'i2p_port': value['port'],
                }
            )
        elif value['type'] != lt.proxy_type_t.none:
            proxy_settings.update(
                {
                    'proxy_type': value['type'],
                    'proxy_hostname': value['hostname'],
                    'proxy_port': value['port'],
                    'proxy_username': value['username'],
                    'proxy_password': value['password'],
                }
            )

        self.core.apply_session_settings(proxy_settings)

    def _on_set_rate_limit_ip_overhead(self, key, value):
        self.core.apply_session_setting('rate_limit_ip_overhead', value)

    def _on_set_geoip_db_location(self, key, geoipdb_path):
        # Load the GeoIP DB for country look-ups if available
        if os.path.exists(geoipdb_path):
            try:
                self.core.geoip_instance = GeoIP(geoipdb_path, 0)
            except Exception as ex:
                log.warning('GeoIP Unavailable: %s', ex)
        else:
            log.warning('Unable to find GeoIP database file: %s', geoipdb_path)

    def _on_set_cache_size(self, key, value):
        self.core.apply_session_setting('cache_size', value)

    def _on_set_cache_expiry(self, key, value):
        self.core.apply_session_setting('cache_expiry', value)

    def _on_auto_manage_prefer_seeds(self, key, value):
        self.core.apply_session_setting('auto_manage_prefer_seeds', value)



#
# Copyright (C) 2008-2009 Ido Abramovich <ido.deluge@gmail.com>
# Copyright (C) 2009 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import json
import logging
import re

import deluge.component as component
import deluge.ui.console.utils.colors as colors
from deluge.ui.client import client

from . import BaseCommand

log = logging.getLogger(__name__)


def json_eval(source):
    """Evaluates string as json data and returns Python objects."""
    if source == '':
        return source

    src = source.splitlines()[0]

    # Substitutions to enable usage of pythonic syntax.
    if src.startswith('(') and src.endswith(')'):
        src = re.sub(r'^\((.*)\)$', r'[\1]', src)
    elif src.lower() in ('true', 'false'):
        src = src.lower()
    elif src.lower() == 'none':
        src = 'null'

    try:
        return json.loads(src)
    except ValueError:
        return src


class Command(BaseCommand):
    """Show and set configuration values"""

    usage = _('Usage: config [--set <key> <value>] [<key> [<key>...] ]')

    def add_arguments(self, parser):
        set_group = parser.add_argument_group('setting a value')
        set_group.add_argument(
            '-s',
            '--set',
            action='store',
            metavar='<key>',
            help=_('set value for this key'),
        )
        set_group.add_argument(
            'values', metavar='<value>', nargs='+', help=_('Value to set')
        )
        get_group = parser.add_argument_group('getting values')
        get_group.add_argument(
            'keys',
            metavar='<keys>',
            nargs='*',
            help=_('one or more keys separated by space'),
        )

    def handle(self, options):
        self.console = component.get('ConsoleUI')
        if options.set:
            return self._set_config(options)
        else:
            return self._get_config(options)

    def _get_config(self, options):
        def _on_get_config(config):
            string = ''
            for key in sorted(config):
                if key not in options.values:
                    continue

                color = '{!white,black,bold!}'
                value = config[key]
                try:
                    color = colors.type_color[type(value)]
                except KeyError:
                    pass

                # We need to format dicts for printing
                if isinstance(value, dict):
                    import pprint

                    value = pprint.pformat(value, 2, 80)
                    new_value = []
                    for line in value.splitlines():
                        new_value.append(f'{color}{line}')
                    value = '\n'.join(new_value)

                string += f'{key}: {color}{value}\n'
            self.console.write(string.strip())

        return client.core.get_config().addCallback(_on_get_config)

    def _set_config(self, options):
        config = component.get('CoreConfig')
        key = options.set
        val = ' '.join(options.values)

        try:
            val = json_eval(val)
        except Exception as ex:
            self.console.write('{!error!}%s' % ex)
            return

        if key not in config:
            self.console.write('{!error!}Invalid key: %s' % key)
            return

        if not isinstance(config[key], type(val)):
            try:
                val = type(config[key])(val)
            except TypeError:
                self.config.write(
                    '{!error!}Configuration value provided has incorrect type.'
                )
                return

        def on_set_config(result):
            self.console.write('{!success!}Configuration value successfully updated.')

        self.console.write(f'Setting "{key}" to: {val!r}')
        return client.core.set_config({key: val}).addCallback(on_set_config)

    def complete(self, text):
        return [k for k in component.get('CoreConfig') if k.startswith(text)]
